import pytest
from acequia import HydroMonitor, GwSeries
from pandas import Series, DataFrame

fpath = r'.\data\hymon\hydromonitor_testdata.csv'

@pytest.fixture
def hm():
    return HydroMonitor(fpath)

def test_repr(hm):
    assert isinstance(hm,HydroMonitor)
    assert isinstance(str(hm),str)


# test pulbic methods
# -------------------

def test_idkeys(hm):
    assert isinstance(hm.idkeys,list)

def test_delete_duplicate_data(hm):
    res = hm._delete_duplicate_data()
    assert not res.empty

def test_get_series(hm):
    gws = hm.get_series(loc='B29A0072',fil='1')    
    assert isinstance(gws,GwSeries)

def test_series(hm):
    assert isinstance(hm.series,list)
    assert len(hm.series)!=0

def test_locations(hm):
    assert isinstance(hm.locations,list)
    assert len(hm.locations)!=0

def test_to_list(hm):
    gwlist = hm.to_list()
    assert isinstance(gwlist,list)
    for i,gw in enumerate(gwlist):
        assert isinstance(gw,GwSeries)

def test_to_json(hm):
    hm.to_json(r'.\output\json\\')

"""
def test_iterdata(hm):
    rownr = 0
    for idx,row in hm.iterdata():
        assert len(idx)!=0
        assert len(row)!=0
        rownr+=1
    assert(rownr>0)
"""

def test_iteritems(hm):
    for gw in hm.iteritems():
        assert isinstance(gw, GwSeries)


# test private methods
# --------------------

def test__readcsv(hm):
    header, line_numbers, meta_colnames, data_colnames = hm._readcsv()
    assert not header.empty
    assert isinstance(line_numbers,Series)
    assert isinstance(meta_colnames,list)
    assert len(meta_colnames)!=0
    assert isinstance(data_colnames,list)
    assert len(data_colnames)!=0

def test__extract_contents(hm):
    metadata, data = hm._extract_contents()
    assert not metadata.empty
    assert not data.empty










import numpy as np
import matplotlib.pyplot as plt
import scipy.stats
import statsmodels.api as sm
import statsmodels.graphics.tsaplots
from statsmodels.tsa.stattools import ccf, acf, pacf


def plot_tsmodel_statistics(obs=None,sim=None,res=None,noise=None,
    figtitle=None,figtype='full'):
    """Plot statistical diagnostics of a time series model

    Parameters
    ----------
    obs : pd.Series
        time series with observations
    sim : pd.Series
        time series with simulated values
    res : pd.Series
        time series with residuals
    noise : pd.Series
        time series with noise
    figtitle : str
        title above output graph

    Notes
    -----
    Definition of input time series is as followed:
    obs : observed values
    sim : values simulated with a transfer-function
    res : calculated differences between obs and sim
    noise : values of res after a noise model is applied

    """
    pms = TsModelStatsPlot(obs=obs,sim=sim,res=res,noise=noise,figtitle=figtitle)
    return pms.plot(figtype=figtype)


class TsModelStatsPlot():
    """Plot statistical diagnostics of a time series model"""

    _clrdict = {
        'obs' : '#f18805',
        'sim' : '#1f2bfb',
        'res' : '#1f2bfb',
        'noise' : '#8a2be2',
        'stats' : '#5e6175', #'#0000ff', #'#9bccf9' #'#ffcf40'
        'bandwidth' : '#fff98e',
        }

    _axtitlefontsize = 18.
    _axtitle_dict = {'fontsize':10.}
    _noiselim = [None,None]
    _markersize = 5.
    _titlefontsize = 7.
    _suptitlefontsize = 12.


    def __init__(self,obs=None,sim=None,res=None,noise=None,figtitle=None):
        """Parameters
        ----------
        obs : pd.Series
            time series with observations
        sim : pd.Series
            time series with simulated values
        res : pd.Series
            time series with residuals
        noise : pd.Series
            time series with noise
        figtitle : str
            title above output graph

        Notes
        -----
        Definition of input time series is as followed:
        obs : observed values
        sim : values simulated with a transfer-function
        res : calculated differences between obs and sim
        noise : values of res after a noise model is applied

        """

        self._obs = obs
        self._sim = sim
        self._res = res
        self._noise = noise
        self._figtitle = figtitle

        self._fig_width = 6.4
        self._fig_height = 5.4


    def plot(self,figtype='full'):
        """Plot figure

        Parameters
        ---------
        figtype : {'full','basic'}, default 'full'
            model statistics figure type

        Return
        ------
        plt.fig

        """

        if figtype=='full':
            self._fullfigure()

        if figtype=='basic':
            self._basicfigure()

        return self._fig


    def __repr__(self):
        return (f'{self.__class__.__name__}')


    def _fullfigure(self):
        """Define figure and gridspec object """

        # define empty figure and gridspec
        self._nrows = 7
        self._ncols = 2

        self._fig = plt.figure(constrained_layout=True, 
            figsize=(self._fig_width,self._fig_height))
        gs = self._fig.add_gridspec(self._nrows,self._ncols)

        # create empty subaxes using gridspec
        self._axs = {}
        self._axs['mlfit'] = self._fig.add_subplot(gs[0:2,:])
        self._axs['residuals'] = self._fig.add_subplot(gs[2,:])
        self._axs['noise'] = self._fig.add_subplot(gs[3,:])
        self._axs['residuals_acf'] = self._fig.add_subplot(gs[4,0])
        self._axs['residuals_pacf'] = self._fig.add_subplot(gs[4,1])
        self._axs['noise_acf'] = self._fig.add_subplot(gs[5,0])
        self._axs['noise_pacf'] = self._fig.add_subplot(gs[5,1])
        self._axs['noise_histogram'] =  self._fig.add_subplot(gs[6,0])
        self._axs['noise_qq'] = self._fig.add_subplot(gs[6,1])

        # plot subplots
        self._plot_modelfit()
        self._plot_residuals()
        self._plot_noise()
        self._plot_residuals_acf()
        self._plot_residuals_pacf()
        self._plot_noise_acf()
        self._plot_noise_pacf()
        self._plot_noise_histogram()
        self._plot_noise_qq()

        self._format_axes(axnames=['residuals_acf','residuals_pacf',
            'noise_acf','noise_pacf'])

        self._fig.suptitle(self._figtitle, fontsize=self._suptitlefontsize,
            fontweight='bold')

        return self._fig


    def _basicfigure(self):
        """Define figure and gridspec object """

        # define empty figure and gridspec
        self._nrows = 5 #7
        self._ncols = 2
        #self._fig_width = 18
        #self._fig_height = 15

        self._fig = plt.figure(constrained_layout=True, 
            figsize=(self._fig_width,self._fig_height))
        gs = self._fig.add_gridspec(self._nrows,self._ncols)

        # create empty subaxes using gridspec
        self._axs = {}
        self._axs['mlfit'] = self._fig.add_subplot(gs[0:2,:])
        ##self._axs['residuals'] = self._fig.add_subplot(gs[2,:])
        self._axs['noise'] = self._fig.add_subplot(gs[2,:])
        ##self._axs['residuals_acf'] = self._fig.add_subplot(gs[4,0])
        ##self._axs['residuals_pacf'] = self._fig.add_subplot(gs[4,1])
        self._axs['noise_acf'] = self._fig.add_subplot(gs[3:5,0])
        ##self._axs['noise_pacf'] = self._fig.add_subplot(gs[5,1])
        self._axs['noise_histogram'] =  self._fig.add_subplot(gs[3:5,1])
        ##self._axs['noise_qq'] = self._fig.add_subplot(gs[6,1])

        # plot subplots
        self._plot_modelfit()
        ##self._plot_residuals()
        self._plot_noise()
        ##self._plot_residuals_acf()
        ##self._plot_residuals_pacf()
        self._plot_noise_acf()
        ##self._plot_noise_pacf()
        ##self._format_axes()
        self._plot_noise_histogram()
        ##self._plot_noise_qq()

        self._format_axes(axnames=['noise_acf'])

        self._fig.suptitle(self._figtitle, fontsize=self._suptitlefontsize,
            fontweight='bold')

        return self._fig


    def _plot_modelfit(self):
        """plot simulations and measurements in one graph"""

        x = self._sim.index.values
        y = self._sim.values
        clrs = self._clrdict['sim']
        self._axs['mlfit'].plot(x,y,c=clrs,label='transfer model')

        x = self._obs.index.values
        y = self._obs.values
        clrs = self._clrdict['obs']
        self._axs['mlfit'].plot(x,y,linestyle='None',marker='o',
            markersize=self._markersize,
            markerfacecolor=clrs,markeredgecolor='None',
            label='observations')

        self._axs['mlfit'].legend(loc='lower right')
        #self._axs['mlfit'].axes.get_xaxis().set_visible(False)


    def _plot_residuals(self):
        """Plot residuals in seperate graph"""

        x = self._res.index.values
        y = self._res.values
        clrs = self._clrdict['sim']
        self._axs['residuals'].plot(x,y,linestyle='None',marker='o',
            markersize=self._markersize,
            markerfacecolor=clrs,markeredgecolor='None',)
            #label='residuals (correlated)')

        self._axs['residuals'].set_title('residuals (correlated)',
            fontdict=self._axtitle_dict)

        ylim = self._axs['residuals'].get_ylim()
        resmax = round(max(abs(ylim[0]),abs(ylim[1])),1)
        self._reslim = [-resmax,resmax]

        self._axs['residuals'].set_ylim(self._reslim)
        self._axs['residuals'].set_title(
            'Residuals of the transfer-model (correlated)',
            self._axtitle_dict)

        self._axs['residuals'].axes.get_xaxis().set_visible(False)


    def _plot_noise(self):
        """Plot model innovations in seperate graph"""

        x = self._noise.index.values
        y = self._noise.values
        clrs = self._clrdict['noise']
        self._axs['noise'].plot(x,y,linestyle='None',marker='o',
            markersize=self._markersize,
            markerfacecolor=clrs,markeredgecolor='None',)
            #label='noise (uncorrelated)')

        self._axs['noise'].set_title('noise (uncorrelated)',
            self._axtitle_dict)


        ylim = self._axs['noise'].get_ylim()
        noisemax = round(max(abs(ylim[0]),abs(ylim[1])),1)
        self._noiselim = [-noisemax,noisemax]
        self._axs['noise'].set_ylim(self._noiselim)
        self._axs['noise'].set_title('Noise (uncorrelated)',
            self._axtitle_dict)

        self._axs['noise'].axes.get_xaxis().set_visible(False)


    def _plot_residuals_acf(self):
        """Plot autocorrelation of the noise""" 
        residuals = self._res
        statsmodels.graphics.tsaplots.plot_acf(residuals, 
            ax=self._axs['residuals_acf'], lags=30, alpha=0.05,
            zero=False) ##, title='Autocorrelation residuals')

        self._axs['residuals_acf'].set_title(
            'Autocorrelation of the residuals',self._axtitle_dict)


    def _plot_residuals_pacf(self):
        """Plot partial autocorrelation of the noise"""
        residuals = self._res
        statsmodels.graphics.tsaplots.plot_pacf(residuals, ax=self._axs['residuals_pacf'], lags=30, alpha=0.05, 
             zero=False, title='Partial autocorrelation residuals')

        self._axs['residuals_pacf'].set_title(
            'Partial autocorrelation of the residuals',self._axtitle_dict)


    def _plot_noise_acf(self):
        """Plot autocorrelation of the innovations"""
        noise = self._noise
        statsmodels.graphics.tsaplots.plot_acf(noise, 
            ax=self._axs['noise_acf'], lags=30, alpha=0.05, 
            zero=False)
        self._axs['noise_acf'].set_title(
            'Autocorrelation of the noise',self._axtitle_dict)


    def _plot_noise_pacf(self):
        """Plot partial autocorrelation of the innovations"""
        noise = self._noise
        statsmodels.graphics.tsaplots.plot_pacf(noise, 
            ax=self._axs['noise_pacf'], lags=30, alpha=0.05, 
            zero=False)
        self._axs['noise_pacf'].set_title(
            'Partial autocorrelation of the noise',self._axtitle_dict)


    def _format_axes(self,axnames=[]):

        for axname in axnames:

            if axname.startswith('residuals'):
                clrs = self._clrdict['res']
            elif axname.startswith('noise'):
                clrs = self._clrdict['noise']
            else:
                clrs = '#ee22cc'

            ax = self._axs[axname]
            ax.get_lines()[1].set_markerfacecolor(clrs) #'#8a2be2')
            ax.get_lines()[1].set_markeredgecolor(clrs)

            # set color of correlation plot bandwith
            art = ax.properties()['default_bbox_extra_artists'][1]
            art.set_facecolor(self._clrdict['bandwidth'])


    def _plot_noise_histogram(self):
        """Plot histogram of innovations"""

        noise = self._noise
        clrs = self._clrdict['noise']

        # empirical distribution
        n, bins, patches = self._axs['noise_histogram'].hist(noise,
            50, density=True, facecolor=clrs, alpha=0.75)

        # theoretical distribution
        x = np.arange(-0.5,0.5,0.01)
        y = scipy.stats.norm.pdf(x,0,np.std(noise))
        clrs = self._clrdict['stats']
        self._axs['noise_histogram'].plot(x,y,color=clrs,linewidth=2)

        self._axs['noise_histogram'].set_xlim(self._noiselim)
        self._axs['noise_histogram'].set_title(
            'Histogram of the noise',self._axtitle_dict)


    def _plot_noise_qq(self):
        """Plot qqplot of innovations"""

        noise = self._noise
        sm.qqplot(noise, scipy.stats.t, fit=True, line="45", ax=self._axs['noise_qq'])
        line1,line2 = self._axs['noise_qq'].get_lines() #line1: innovations line2: normal distribution
        line1.set_markerfacecolor('#8a2be2')
        line1.set_markeredgecolor('None')
        line1.set_markersize(4)
        line2.set_color(self._clrdict['stats']) #'#666666')
        #axs[2,1].set_xlabel('Theoretische kwantielen standaard normale verdeling',fontsize=12)
        #axs[2,1].set_ylabel('Kwantielen noise',fontsize=12)
        self._axs['noise_qq'].set_title(
            'QQ-plot of the noise',self._axtitle_dict)





import pytest
import collections
import numpy as np
from pandas import DataFrame
import pandas as pd
from acequia import GwFiles, GwSeries
import acequia as aq

dinodir = '.\\data\\dinogws\\'
jsondir = '.\\output\\json\\'
csvdir = '.\\output\\csv\\'

@pytest.fixture
def gwf():
    return GwFiles.from_dinocsv(dinodir)

@pytest.fixture
def gwf_json():
    return GwFiles.from_json(jsondir)

def test_from_dinocsv_with_only_filedir():
    gwf = GwFiles.from_dinocsv(dinodir)
    assert isinstance(gwf.filetable,DataFrame)
    assert not gwf.filetable.empty

def test_from_dinocsv_with_loclist(gwf):
    loclist = gwf.filetable['loc'].values

    gwf2 = GwFiles.from_dinocsv(dinodir,loclist=loclist)
    assert isinstance(gwf2.filetable,DataFrame)
    assert len(gwf2.filetable)!=0

def test_from_json_with_only_filedir():

    gwf = GwFiles.from_json(jsondir)
    assert isinstance(gwf.filetable,DataFrame)
    assert not gwf.filetable.empty

def test_from_json_with_loclist(gwf):

    loclist = gwf.filetable['loc'].values ##[:listlen]

    gwf2 = GwFiles.from_json(jsondir,loclist=loclist)
    assert isinstance(gwf2.filetable,DataFrame)
    assert len(gwf2.filetable)!=0

def test_init_with_invalid_input():

    with pytest.raises(ValueError):
        gwf = GwFiles('A_String')

    with pytest.raises(ValueError):
        badcolumns = DataFrame(columns=['A','B','C',])
        gwf = GwFiles(badcolumns)

def test_repr(gwf):

    assert isinstance(repr(gwf),str)

def test_iteritems(gwf, gwf_json):

    for gw in gwf.iteritems():
        assert isinstance(gw,GwSeries)

    for gw in gwf_json.iteritems():
        assert isinstance(gw,GwSeries)


def test_to_json(gwf):

    jsn = gwf.to_json(jsondir)
    assert isinstance(jsn,list)
    assert isinstance(jsn[0],collections.OrderedDict)

def test_to_csv(gwf):

    csv = gwf.to_csv(csvdir)
    assert isinstance(csv,list)
    assert isinstance(csv[0],pd.Series)





import pytest
import collections
from pandas import DataFrame
from acequia.read.gpxtree import GpxTree

fpath = r'.\data\gpx\BrittanyJura\Ouistreham_Caen.gpx'

# test constructors
def test_init_raises_error_with_None_as_parameter():
    with pytest.raises(AttributeError):
        GpxTree(None)

def test_classmethod_read_from_file():
    gpxtree = GpxTree.from_file(fpath)
    assert isinstance(gpxtree,GpxTree)

@pytest.fixture
def gpxtree():
    """Return GpxTree instance with tree from gpx file"""
    return GpxTree.from_file(fpath)

# test private methods
def test__get_meta(gpxtree):
    result = gpxtree._get_meta()
    assert isinstance(result,collections.abc.Mapping)

def test__get_waypoints(gpxtree):
    result = gpxtree._get_waypoints()
    assert isinstance(result,DataFrame)
    assert not result.empty

def test__get_trackpoints(gpxtree):
    result = gpxtree._get_trackpoints()
    assert isinstance(result,DataFrame)
    assert not result.empty

# test properties
def test_meta(gpxtree):
    assert isinstance(gpxtree.meta, collections.abc.Mapping)

def test_bounds(gpxtree):
    assert isinstance(gpxtree.bounds, collections.abc.Mapping)

def test_trackpoints(gpxtree):
    assert isinstance(gpxtree.trackpoints,DataFrame)
    assert not gpxtree.trackpoints.empty

def test_waypoints(gpxtree):
    assert isinstance(gpxtree.waypoints,DataFrame)
    assert not gpxtree.waypoints.empty

def test_contents(gpxtree):
    assert isinstance((list(gpxtree.contents)),list)




""" This module contains a class GwGxg that calculates some
descriptive statistics from a series of groundwater head measurements
used by groundwater practitioners in the Netherlands 

History: Created 16-08-2015, last updated 12-02-1016
         Migrated to acequia on 15-06-2019

@author: Thomas de Meij

"""

import math
from datetime import datetime
import datetime as dt
import warnings
import numpy as np
from pandas import Series, DataFrame
import pandas as pd

from .. import gwseries
from .utils import ts1428, hydroyear, season


def stats_gxg(ts, reference='datum', minimal=False, surface=None):
    """Return table with GxG statistics

    Parameters
    ----------
    ts : aq.GwSeries, pd.Series
        Groundwater head time series
    reference : {'datum','surface'}, optional
        Reference level for groundwater heads
    minimal : bool, default False
        Return minimal number of gxg statistics.
    surface : float, default None
        Surface level.

    Returns
    -------
    pd.DataFrame

    """
    gxg = GxgStats(ts, surface=surface)
    return gxg.gxg(reference=reference, minimal=minimal)


class GxgStats:
    """Calculate descriptive statistics for time series of measured heads

    Parameters
    ----------
    gw : aq.GwSeries, pd.Series
        timeseries of groundwater head measurements relative to datum level

    srname : str, optional
        name of groundwater head series

    surface : float, optional
        surface level height (if ref='datum' this option is ignored)

    Notes
    -----
    In the Netherlands, traditionally groundwater head series are 
    summarized using decriptive statistics that characterise the mean 
    highest level (GHG), the mean lowest level (GLG) and the mean spring 
    level (GVG). These three measures together are reffered to as the GxG.
    The definitions of GHG, GLG and GVG are based on time series with 
    measured heads on the 14th and 28th of each month. Therefore the time 
    series of measrued heads is internally resampled to values on the 14th
    and 28yh before calculating the GxG statistics.

    For further reference: 
    P. van der SLuijs and J.J. de Gruijter (1985). 'Water table classes: 
    a method to decribe seasonal fluctuation and duration of water table 
    classes on Dutch soil maps.' Agricultural Water Management 10 (1985) 
    109 - 125. Elsevier Science Publishers, Amsterdam.

    """

    N14 = 18
    ## REFERENCE = ['datum','surface']
    APPROXIMATIONS = ['SLUIJS82','HEESEN74','SLUIJS76a','SLUIJS76b',
        'SLUIJS89pol','SLUIJS89sto','RUNHAAR89','GAAST06',]
    VGDATES = ['apr1','apr15','mar15']
    VGREFDATE = 'apr1'


    def __init__(self, gw, srname=None, surface=None):
        """Return GxG object"""

        if isinstance(gw,gwseries.GwSeries):

            self._ts = gw.heads(ref='datum')
            self.srname = gw.name()
            if surface is None:
                self._surface = gw.surface()
            else:
                self._surface = surface
            self._gw = gw

        elif isinstance(gw,pd.Series):

            self._ts = gw
            self.srname = self._ts.name
            if surface is not None:
                self._surface = surface
            else:
                self._surface = np.nan
            self._gw = None

        else:
            raise(f'{gw} is not of type GwSeries or pd.Series')

        self._ts1428 = ts1428(self._ts,maxlag=3,remove_nans=False)
        self._xgnap = self._calculate_xg_nap()


    def _yearseries(self,ts,dtype='float64'):
        """Return empty time series with years as index with all years
        between min(year) and max(year) in index (no missing years)"""

        if isinstance(ts,pd.Series):
            years = set(ts.index.year)

        elif isinstance(ts,(list,set,np.ndarray)):
            years = set(ts)

        else:
            raise(f'{ts} must be list-like')

        minyear = min(years)
        maxyear= max(years)
        sr = Series(index=range(minyear,maxyear+1),dtype=dtype,name='year')
        return sr


    def vg3(self):
        """Return VG3 (Spring Level) for each year

        VG3 is calculated as the mean of groundwater head 
        levels on 14 march, 28 march and 14 april

        Return
        ------
        pd.Series


        Notes
        -----
        Calculation of GVG based on the average of three dates was 
        introduced by Finke et al. (1999)

        References
        ----------
        Finke, P.A., D.J. Brus, T. Hoogland, J. Oude Voshaar, F. de Vries
        & D. Walvoort (1999). Actuele grondwaterinformatie 1:10.000 in de
        waterschappen Wold en Wieden en Meppelerdiep. Gebruik van digitale 
        maaiveldshoogtes bij de kartering van GHG, GVG en GLG. SC-rapport
        633. (in Dutch).
        """

        self._vg3 = self._yearseries(self._ts1428)
        for i,year in enumerate(self._vg3.index):

            v1 = self._ts1428[dt.datetime(year,3,14)]
            v2 = self._ts1428[dt.datetime(year,3,28)]
            v3 = self._ts1428[dt.datetime(year,4,14)]

            with warnings.catch_warnings():
                # numpy raises a silly warning with nanmean on NaNs
                warnings.filterwarnings(action='ignore', 
                    message='Mean of empty slice')
                self._vg3[year] = np.round(np.nanmean([v1,v2,v3]),2)

        self._vg3.name = 'VG3'
        return self._vg3


    def vg1(self,refdate=VGREFDATE,maxlag=7):
        """Return VG (Spring Level) for each year as the measurement
        closest to refdate

        Parameters
        ----------
        refdate : {'apr1','apr15','mar15'}, default 'apr1'
            reference date for estimating VG

        maxlag : number
            maximum allowed difference between measurement date en refdate


        Return
        ------
        pd.Series 

        Notes
        -----
        The VG (Voorjaars  Grondwaterstand, Spring Level) is estimated as 
        the single measurement closest to the reference date given by 
        refdate.

        The reference date for calculation of the GVG was changed from
        april 15 to april 1st in de early eighties. In 2000 the 
        Cultuurtechnisch Vademecum proposed march 15 as the new reference 
        date for the GVG but this proposal was not generally adopted.
        In practice april 1st is allways used as reference date and this 
        is used as default for calculations.

        References
        ----------
        Van der Gaast, J.W.J., H.Th.L. Massop & H.R.J. Vroon (2009). Actuele
        grondwaterstandsituatie in natuurgebieden. Rapport 94 WOT. Alterra,
        Wageningen. (in Dutch).
        """

        if refdate not in self.VGDATES:
            warnings.warn((f'Reference date {refdate} for GVG is not '
                f'recognised. Reference date \'{self.VGREFDATE}\' is '
                f'assumed.'))
            refdate = self.VGREFDATE

        vg1 = self._yearseries(self._ts1428)
        for i,year in enumerate(vg1.index):

            if refdate=='apr1':
                date = dt.datetime(year,4,1)
            if refdate=='apr15':
                date = dt.datetime(year,4,15)
            if refdate=='mar15':
                date = dt.datetime(year,3,15)

            daydeltas = self._ts.index - date
            mindelta = np.amin(np.abs(daydeltas))
            sr_nearest = self._ts[np.abs(daydeltas) == mindelta]

            maxdelta = pd.to_timedelta(f'{maxlag} days')
            if (mindelta <= maxdelta):
                vg1[year] = np.round(sr_nearest.iloc[0],2)

        vg1.name = f'VG{refdate}'
        return vg1


    def _calculate_xg_nap(self):
        """Calculate xg statistics for eacht year and return table""" 

        hydroyears = hydroyear(self._ts1428)
        sr = self._yearseries(hydroyears)
        xg = pd.DataFrame(index=sr.index)
        xg.index.name = 'year'

        for year in xg.index:

            ts = self._ts1428[hydroyears==year]
            ts = ts[ts.notnull()]

            n1428 = len(ts)
            if not np.isnan(n1428):
                n1428 = math.floor(n1428)

            hg3 = np.nan
            lg3 = np.nan

            if n1428 >= self.N14:

                hg3 = ts.nlargest(n=3).mean()
                lg3 = ts.nsmallest(n=3).mean()

            hg3w = np.nan
            lg3s = np.nan

            if n1428 >= self.N14:

                ts_win = ts[season(ts)=='winter']
                ts_sum = ts[season(ts)=='summer']

                hg3w = ts_win.nlargest(n=3).mean()
                lg3s = ts_sum.nsmallest(n=3).mean()

            xg.loc[year,'hg3'] = np.round(hg3,2)
            xg.loc[year,'lg3'] = np.round(lg3,2)
            xg.loc[year,'hg3w'] = np.round(hg3w,2)
            xg.loc[year,'lg3s'] = np.round(lg3s,2)
            xg['vg3'] = self.vg3()

            for date in self.VGDATES:
                xg[f'vg_{date}'] = self.vg1(refdate=date)

            xg.loc[year,'n1428'] = n1428

        return xg


    def xg(self, reference='datum', name=True):
        """Return table of GxG groundwater statistics for each 
        hydrological year

        Parameters
        ----------
        reference : {'datum','surface'}, default 'datum'
            Reference level for gxg statistics.
        name : bool, default True
            Include series name in index.

        Returns
        -------
        pd.DataFrame

        Notes
        -----
        Results are given in meter for reference level 'datum' and in 
        centimeter for reeference level 'surface'.
        ..."""

        if reference not in ['datum','surface']:
            warnings.warn((f'Unknown reference level "{reference}". '
                f'Reference level "datum" is assumed.'))
            reference = 'datum'

        if reference == 'datum':
            multiplier = 1
        elif reference == 'surface':
            multiplier = 100

        xg = self._xgnap.copy()
        if name==True:
            xg = pd.concat({self.srname: xg}, names=['series'])

        if reference=='surface':
            # convert unit meter to centimeter
            for col in xg.columns:

                if col in ['n1428']:
                    continue

                xg[col] = (self._surface - xg[col])*multiplier
                xg[col] = xg[col].apply(lambda x:math.floor(x) if 
                    not np.isnan(x) else x)

                ##if not np.isnan(xg[col]):
                ##    xg[col] = math.floor(xg[col])

        return xg


    def gxg(self, reference='datum', minimal=False):
        """Return table with GxG for one head series

        Parameters
        ----------
        minimal : bool, default True
            return minimal selection of stats
        reference : {'datum','surface'}, default 'datum'
            reference level for gxg statistics

        Returns
        -------
        pd.DataFrame"""

        """
        if hasattr(self,'_minimal'):
            if self._minimal!=minimal:
                self._reset()
        self._minimal = minimal

        if self._reflev==reflev:
            if hasattr(self,'_gxg'):
                return self._gxg
        else:
            self._reset()
            self._validate_reflev (reflev)
        """

        xg = self.xg(reference=reference,name=False)

        gxg = pd.Series(name=self.srname,dtype='object')
        for col in xg.columns:
            sr = xg[col][xg[col].notnull()]

            if reference=='datum':
                gxg[col] = np.round(sr.mean(),2)

            if reference=='surface':
                ##gxg[col] = np.round(sr.mean())
                if not np.isnan(sr.mean()):
                    gxg[col] = math.floor(sr.mean())
                else:
                    gxg[col] = np.nan

            if col=='n1428':
                gxg[col] = math.floor(sr.mean())

        # calculate gt
        gxg['gt'] = self.gt()

        gxg['gxgref'] = reference

        # calculate std
        for col in xg.columns:

            if col in ['n1428',]: #'measfrq']:
                continue

            if reference=='datum':
                gxg[col+'_std'] = np.round(xg[col].std(
                    skipna=True),2)

            elif reference=='surface':
                sr = xg[col]
                gxg[col+'_std'] = np.round(sr.std(skipna=True))

            else:
                raise ValueError((f'Reference level {reference} is not valid.',
                    f'Valid reference levels are \'datum\' or \'surface\'.'))

        # calculate standard error
        for col in xg.columns:

            if col in ['n1428',]:
                continue

            if reference=='datum':
                sr = xg[col]
                gxg[col+'_se'] = np.round(sr.std(skipna=True
                    )/np.sqrt(sr.count()),2)

            if reference=='surface':
                sr = xg[col]
                gxg[col+'_se'] = np.round(sr.std(skipna=True
                    )/np.sqrt(sr.count()),0)

        # count nyears
        for col in xg.columns:

            if col in ['n1428',]:
                continue

            sr = xg[col][xg[col].notnull()]
            gxg[f'{col}_nyrs'] = np.round(sr.count())

        replacements = [('hg3','ghg'),('lg3','glg'),('vg','gvg'),]
        for old,new in replacements:
            gxg.index = gxg.index.str.replace(old,new)

        # gvg approximation formulas
        if reference=='surface':
            for apx in self.APPROXIMATIONS:
                rowname = 'gvg_'+apx.lower()
                gxg[rowname] = self.gvg_approximate(apx)

        self._gxg = gxg        

        if minimal:
            colnames = ['ghg','glg','gvg3','gvg_apr1','gt','gxgref',
                'n1428',]
            gxg = gxg[gxg.index.intersection(colnames)]

        return gxg


    def ghg(self):
        """Return mean highest level (GHG)"""

        if not hasattr(self,'_gxg'):
            self.gxg()

        return self._gxg['ghg']


    def glg(self):
        """Return mean highest level (GHG)"""

        if not hasattr(self,'_gxg'):
            self.gxg()

        return self._gxg['glg']


    def gt(self):
        """Return groundwater class table as str"""

        if not hasattr(self,'_xg'):
            self._calculate_xg_nap()

        # do not call self._gxg to avoid recursion error because gt() 
        # is used in gxg()

        with warnings.catch_warnings():
            # numpy raises a silly warning with nanmean on NaNs
            warnings.filterwarnings(action='ignore', 
                message='Mean of empty slice')

            ghg = (self._surface - np.nanmean(self._xgnap['hg3']))*100
            glg = (self._surface - np.nanmean(self._xgnap['lg3']))*100

        if (ghg<20) & (glg<50):
            return 'I'

        if (ghg<25) & (50<glg<80):
            return 'II'

        if (25<ghg<40) & (50<glg<80):
            return 'II*'

        if (ghg<25) & (80<glg<120):
            return 'III'

        if (25<ghg<40) & (80<glg<120):
            return 'III*'

        if (ghg>40) & (80<glg<120):
            return 'IV'

        if (ghg<25) & (glg>120):
            return 'V'

        if (25<ghg<40) & (glg>120):
            return 'V*'

        if (40<ghg<80) & (glg>120):
            return 'VI'

        if (80<ghg<140):
            return 'VII'

        if (ghg>140):
            return 'VII*'

        return np.nan
        # acer palmatum


    def gvg_approximate(self,formula=None):
        """Return GVG calculated with approximation based on GHG and GLG

        Parameters
        ----------
        formula : {'VDS82','VDS89pol','VDS89sto','RUNHAAR'}, default 'VDS82'

        Notes
        -----
        Values for GHG and GLG can be estimated from visual soil profile
        characteristics, allowing mapping of groundwater classes on soil
        maps. GVG unfortunately can not be estimeted is this way.
        Therefore, several regression formulas have been given in litera-
        ture for estimating GVG from GHG and GLG estimates. Three of them
        are implemented: Van der Sluijs (1982), Van der Sluijs (1989) and
        Runhaar (1989)"""


        if formula is None:
            formula = self.APPROXIMATIONS[0]

        if formula not in self.APPROXIMATIONS:
            warnings.warn(f'GVG approximation formula name {formula} not'
                f'recognised. {self.APPROXIMATIONS[0]} is assumed.')

        if not hasattr(self,'_xgnap'):
            self._calculate_xg_nap()

        if formula in ['SLUIS89pol','SLUIS89sto']:
            with warnings.catch_warnings():
                # numpy raises a silly warning with nanmean on NaNs
                warnings.filterwarnings(action='ignore', 
                    message='Mean of empty slice')

                GHG = np.nanmean(self._xgnap['hg3w'])
                GLG = np.nanmean(self._xgnap['lg3s'])

        else:
            with warnings.catch_warnings():
                # numpy raises a silly warning with nanmean on NaNs
                warnings.filterwarnings(action='ignore', 
                    message='Mean of empty slice')

                GHG = np.nanmean(self._xgnap['hg3'])
                GLG = np.nanmean(self._xgnap['lg3'])

        GHG = (self._surface-GHG)*100
        GLG = (self._surface-GLG)*100

        if formula=='HEESEN74': # april 15th
            GVG = 0.2*(GLG-GHG)+GHG+12

        elif formula=='SLUIJS76a': # april 14th
            GVG = 0.15*(GLG-GHG)+(1.01*GHG)+14.3

        elif formula=='SLUIJS76b': # april 14th
            GVG = 1.03*GHG+27.3

        elif formula=='SLUIJS82': 
            GVG = 5.4 + 1.02*GHG + 0.19*(GLG-GHG)

        elif formula=='RUNHAAR89':
            GVG = 0.5 + 0.85*GHG + 0.20*GLG # (+/-7,5cm)

        elif formula=='SLUIJS89pol':
            GVG = 12.0 + 0.96*GHG + 0.17*(GLG-GHG)

        elif formula=='SLUIJS89sto':
            GVG = 4.0 + 0.97*GHG + 0.15*(GLG-GHG)

        elif formula=='GAAST06':
            GVG = 13.7 + 0.70*GHG + 0.25*GLG

        else:
            raise ValueError((f'\'{formula}\' was not recognised as a gvg '
                f'approximation formula. Valid names are '
                f'{self.APPROXIMATIONS}'))

        if not np.isnan(GVG):
            GVG = math.floor(GVG)

        return GVG




import numpy as np
from pandas import Series, DataFrame
import pandas as pd

from ..gwseries import GwSeries


class Dawaco:
    """Manage DAWACO ground water heads data set.

    Constructor
    -----------
    from_excel
        Read DAWACO hydroseries xlsx export file and return Dawaco object.
    ...
    """
    _columns_mapping = {
        'Meetpuntcode' : 'locname',
        'X-coor.(m)' : 'xcr',
        'Y-coor.(m)' : 'ycr',
        'Maaiveld' : 'surfacelevel',
        'Startdatum' : 'startdate',
        'Bkb (m NAP)':'mplevel',
        'Bk Filt' : 'filtop',
        'Ok Filt' : 'filbot',
        'Filter' : 'filname',
        'Datum_Tijd' : 'datetime',
        'Waarde (m -NAP)' : 'meas_nap',
        'Betrouwbaarheid' : 'reliabilty',
        }

    def __init__(self,rawdata=None,fpath=None,title=None):

        self.rawdata = rawdata
        self.fpath = fpath
        self.title = title

        self.rawdata = self.rawdata.rename(
            columns=self._columns_mapping, errors="raise")

    def __len__(self):
        seriescount = len(self.rawdata[['locname','filname']].drop_duplicates())
        return seriescount

    def __repr__(self):
        return f'{self.__class__.__name__} (n={len(self)})'

    @classmethod
    def from_excel(cls, fpath, title=None):
        """Import Dawaco dataset from Excel export format."""
        rawdata = pd.read_excel(fpath,parse_dates=[['Datum','Tijd']])
        rawdata['Filter'] = rawdata['Filter'].astype(str)
        return cls(rawdata=rawdata, fpath=fpath, title=title)

    @property
    def filters(self):
        """Return list of tuples with (locname, filter) for all filters."""
        filters = self.rawdata[['locname','filname']].drop_duplicates()
        return list(zip(filters['locname'],filters['filname']))

    def get_gwseries(self,loc,fil):
        """Return data as list of Acequia GwSeries instances."""

        tbl = self.rawdata
        gw = GwSeries()

        # locprops
        idx = self.rawdata.index[0]
        gw._locprops['locname'] = loc
        gw._locprops['filname'] = fil
        gw._locprops['alias'] = np.nan
        gw._locprops['xcr'] = self.rawdata.loc[idx,'xcr']
        gw._locprops['ycr'] = self.rawdata.loc[idx,'ycr']
        gw._locprops['height_datum'] = 'NAP'
        gw._locprops['grid_reference'] = 'RD'

        # tubeprops
        tp = self.rawdata.copy()
        for col in gw._tubeprops_names:
            if col not in tp.columns:
                tp[col] = np.nan
        tp = tp[gw._tubeprops_names].drop_duplicates(keep='first',ignore_index=True)
        gw._tubeprops = tp.copy()

        # heads
        heads = gw._heads
        heads['headdatetime'] = self.rawdata['datetime']
        heads['headmp'] = self.rawdata['mplevel'] - self.rawdata['meas_nap']
        gw._heads = heads.reset_index(drop=True)

        return gw

    def iteritems(self):
        """Iterate over all groundwater head series and return gwseries objects."""

        for (loc,fil),tbl in self.rawdata.groupby(['locname','filname']):
            gw = self.get_gwseries(loc,fil)
            yield gw






import warnings
import numpy as np
from pandas import DataFrame,Series
import pandas as pd
from shapely.geometry import Point
from geopandas import GeoDataFrame



def pointshape_write(tbl=None,xfield=None,yfield=None,
    filepath='shapefile.shp'):
    """Write pandas dataframe to point shapefile

    Parameters
    ----------
    tbl : pd.DataFrame
        table with data
    xfield : str
        column name with xcoordinates
    yfield : str
        column name with ycoordinates
    filename : str
        shapefile name

    """
    psw = PointShapeWriter(tbl=tbl,xfield=xfield,yfield=yfield,
        filepath=filepath)


class PointShapeWriter:
    """Write dataset to shapefile

    Methods
    -------
    
    """

    def __init__(self,tbl=None,xfield=None,yfield=None,
        filepath='shapefile,shp'):
        """Write pandas dataframe to point shapefile

        Parameters
        ----------
        tbl : pd.DataFrame
            table with data
        xfield : str
            column name with xcoordinates
        yfield : str
            column name with ycoordinates
        filename : str
            shapefile name

        """
        self.tbl = tbl.copy()
        self.xfield = xfield
        self.yfield = yfield
        self.filepath = filepath

        self.indexname = self.tbl.index.name
        #self.colnames = [x for x in self.tbl.columns 
        #    if x not in [self.xfield,self.yfield]]
        self.colnames = self.tbl.columns

        self._remove_missing_crd()
        self._create_shapefile()
        #self.write()

    @classmethod
    def writefile(cls,tbl=None,xfield=None,yfield=None,
        filepath='shapefile,shp'):

        shapewriter = cls.__init__(cls,tbl=tbl,xfield=xfield,
        yfield=yfield,filepath=filepath)
        shapewriter.write()
        return shapewriter


    def _remove_missing_crd(self):

        # convert coordinates to numeric and delete rows with missing coords
        self.tbl[self.xfield] = pd.to_numeric(self.tbl[self.xfield])
        self.tbl[self.yfield] = pd.to_numeric(self.tbl[self.yfield])

        # remove points with missing coordinates
        xmask = self.tbl[self.xfield].notnull().values
        ymask = self.tbl[self.yfield].notnull().values
        nmiss = len(self.tbl[~xmask | ~ymask])
        self.tbl = self.tbl[xmask & ymask].copy()

        # show warning
        if nmiss!=0:
            msg = f'{nmiss} rows with missing coordinates were deleted.'
            warnings.warn(msg)


    def _create_shapefile(self):

        wp = self.tbl
        geometry = [Point(xy) for xy in zip(wp['xcr'],wp['ycr'])]
        self.gdf = GeoDataFrame(wp, geometry=geometry)
        self.gdf = self.gdf.set_crs('epsg:7415')

        for colname in self.gdf.columns:
            if colname not in ['geometry']:
                self.gdf[colname] = self.gdf[colname].astype(str)


    def write(self):

        """
        # create scheme with all cols as str
        propdict = {}
        propdict[self.gdf.index.name] = 'str'
        for colname in self.gdf.columns:
            propdict[colname] = 'str'
        propdict.pop('geometry', None)

        schema = {
            'geometry': 'Point',
            'properties': propdict
            }
        """

        # write shapefile
        self.gdf.to_file(self.filepath) #,schema=schema)



""" 
This module contains the class HeadsDiff for calculating and plotting 
differences between multiple groundwater head series.
"""

import warnings
import math
import numpy as np
from pandas import Series, DataFrame
from pandas import DatetimeIndex
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib as mpl
import seaborn as sns; sns.set()

from .gwseries import GwSeries



class HeadsDif:
    """Calculate head differences and descriptive statistics for 
    multiple groundwater head series

    Methods
    -------
    table_headsdif
        Return table with head differences relative to series ref.
    table_headsref
        Return table with heads relative to mean of entire series.
    difsums
        Return table with head differences by season.
    date_seasons
        Return list with seasons for each datetime in index.
    plot_time
        Plot al heads in one graph and all head differences below that 
        in one figure.
    plot_head
        Plot head difference by reference head value.
    plot_freq
        Plot head differences as grid of frequency plots.

    Notes
    -----
    Class HeadsDif requires input of a list of multiple GwSeries objects
    or pd.Series objects with valid head data. No checking is performed.

    To check input data before creating a HeadsDif object, us the custom
    function headsdif_from_gwseries(). This function returns None when 
    no valid heads could be calculated. 

    Examples
    --------
    Create headdiff object without checking input:
    >>>hd = acequia.headsdif(heads=<list>,locname=<str>, refcol=<str>)

    Call function headsdif_from_gwseires() and clean up data before
    creating a headsdiff object:
    >>>hd = headsdif_from_gwseries(heads=<heads>,locname=<locname>,
            refcol=<refcol>)
    >>>if hd is None:
    >>>     continue
    """


    FOURSEASONSKEY = 'seasons'
    FOURSEASONS = ['Winter','Winter','Spring','Spring',
        'Summer','Summer','Summer','Summer','Autumn', 
        'Autumn','Autumn', 'Winter',]
    FOURSEASONS_COLORS =  {'Winter':'#0057e7', 'Spring':'#74d4f0',
        'Summer':'#ff0000','Autumn':'#ffa500'}
    FOURSEASONS_COLNAMES = ['Spring','Summer','Autumn','Winter']
    TWOSEASONSKEY = 'biannual'
    TWOSEASONS = ['Winter', 'Winter', 'Winter', 'Summer', 'Summer', 
        'Summer', 'Summer', 'Summer', 'Summer', 'Winter', 'Winter', 
        'Winter']
    TWOSEASONS_COLORS = {'Winter':'#2196f3','Summer':'#ffb74d'}
    PERIOD_DEFAULT = FOURSEASONSKEY
    PERIOD_NAMES = {FOURSEASONSKEY,TWOSEASONSKEY}

    MPL_STYLE_DEFAULT = 'seaborn-v0_8-darkgrid'


    def __init__(self,heads=None,locname=None,refcol=None):
        """
        Parameters
        ----------
        heads : pd.DataFrame
            Measured heads of multiple groundwater series.
        locname : str, optional
            Location name for annotating graphs.
        refcol : str, optional
            Series name to use as reference.

        Notes
        -----
        If no values are given for parameters locname and refcol, 
        values will be derived from the first column in heads.
        """

        # validate heads
        if not isinstance(heads,DataFrame):
            raise ValueError((f'Init {self.__class__.__name__}: '
                f'Parameter heads must be a Pandas DataFrame, not '
                f'a {heads.__class__.__name__}.'))
        self.heads = heads

        # validate locname
        if locname is None:
            locname = list(self.heads)[0]
        if '_' in locname:
            locname = locname.split('_')[0]
        self.locname = locname

        # drop empty columns
        for col in self.heads.columns:
            if self.heads[col].isnull().all():
                self.heads = self.heads.drop(columns=[col])

        if len(list(self.heads)) < 2:
            msg = f'Less than two columns with valid data in {self.locname}'
            raise ValueError(msg)

        # validate refcol
        if refcol is None:
            refcol = list(self.heads)[0]
        self.refcol = refcol

        if self.refcol not in list(self.heads):
            old_refcol = self.refcol
            self.refcol = list(self.heads)[0]
            warnings.warn((f'Dropped reference column {old_refcol}, '
                f'replaced with {self.refcol}'))


    def __repr__(self):
        nsr = len(self.heads.columns)
        return (f'{self.__class__.__name__}({nsr} series)')


    @classmethod
    def from_series(cls,heads=None,locname=None,refcol=None):
        """Create HeadsDif object from list of GwSeries objects.

        Parameters
        ----------
        heads : list of GwSeries objects
            Measured heads of multiple groundwater series.
        locname : str, optional
            Location name for annotating graphs, optional.
        refcol : str, optional
            Series name to use as reference.

        Returns
        -------
        HeadsDiff
            HeadsDif object

        Examples
        --------
        >>>hd = HeadsDif.from_gwseries(heads=<heads>,locname=<locname>,
                refcol=<refcol>)

        Notes
        -----
        If no values are given for paramaters locname and refcol, 
        values will be used from the first GwSeries or Series in heads.
        """

        is_no_list_warning = ((f'Parameter heads must be a list of '
            f'GwSeries or Pandas Series objects'))

        if not isinstance(heads,list):
            raise ValueError(is_no_list_warning)

        if isinstance(heads[0],GwSeries):

            if not all([isinstance(x,GwSeries) for x in heads]):
                raise ValueError(is_no_list_warning)

            ##tslist = [gw.heads(ref='datum',freq='D') for gw in heads]
            if locname is None:
                locname = heads[0].locname()
            if refcol is None:
                refcol = heads[0].name()
            heads = [gw.heads() for gw in heads]
            """
            valid_heads = [x for x in heads if not x.heads().isnull().all()]
            if not len(valid_heads)==len(heads):
                removed = list(set(heads).difference(valid_heads))
                warnings.warn((f'Removed following heads series with only '
                    f'NaN valuess: {removed}.'))

            if len(valid_heads)<2:
                msg = 'Less than two heads series left after validation.'
                warnings.warn(msg)
                return None
            """

        if not all([isinstance(x,Series) for x in heads]):
            warnings.warn(is_no_list_warning)
            return None

        headstable = pd.concat(heads, axis=1)
        return HeadsDif(heads=headstable,locname=locname,refcol=refcol)


    def get_relative_heads(self):
        """Return table with heads relative to mean of entire series."""
        relheads = self.heads.copy()
        for i,col in enumerate(list(self.heads)):
            relheads[col] = self.heads[col]-self.heads[col].mean()
        return relheads


    def get_difference(self,refcol=None):
        """Return table with head differences.

        Parameters
        ----------
        refcol : str, optional
            Head differences will be calculated relative to this column.

        Returns
        -------
        pd.DataFrame
        """
        if refcol is None:
            refcol = self.refcol

        if (refcol not in self.heads.columns):
            firstcol = list(self.heads)[0]
            warnings.warn((f'Reference column {refcol} not in heads '
                f'table. First column "{firstcol}" will be used.'))
            refcol = firstcol

        hdif = self.heads.copy()
        for i,col in enumerate(list(hdif)):
            hdif[col] = self.heads[col]-self.heads[refcol]
        return hdif


    def get_seasons(self,dates=None,period=PERIOD_DEFAULT):
        """
        Return list with seasons for each datetime in index.

        Parameters
        ----------
        dates : pd.DataFrame, pd.DateTimeIndex, optional
            Index with dates.

        period : {'seasons','biannual'}, default 'seasons'
            Aggregate by seasons or by half-year.

        Returns
        -------
        numpy array
            Season for each datetime in dates
        """
        if dates is None:
            dates = self.heads

        if isinstance(dates,pd.DataFrame):
            dtindex = dates.index

        if isinstance(dates,pd.Series):
            dtindex = dates.index

        if isinstance(dates,DatetimeIndex):
            dtindex = dates

        if period not in self.PERIOD_NAMES:
            warnings.warn((f"Period '{period}' not in {self.PERIOD_NAMES},"
                "Default {self.PERIOD_DEFAULT} will be used."))
            period = self.PERIOD_DEFAULT

        if period==self.FOURSEASONSKEY:
            month_to_season = dict(zip(range(1,13), self.FOURSEASONS))
            seasons = dtindex.month.map(month_to_season).values

        if period==self.TWOSEASONSKEY:

            month_to_season = dict(zip(range(1,13), self.TWOSEASONS))
            seasons = dtindex.month.map(month_to_season).values

        return seasons


    def get_difference_by_season(self,period=PERIOD_DEFAULT):
        """Return table with head differences grouped by season.

        Parameters
        ----------
        seasons : {'seasons','biannual'}, default 'seasons'
            Aggregate by season or by half-year.

        Returns
        -------
        pd.DataFrame
        """
        headsdif = self.get_difference()
        seasons = self.get_seasons(headsdif,period=period)
        difsns = (headsdif*100).groupby(seasons).mean().round(2).T
        difsns.columns.name = None
        difsns.index.name = 'series'
        if period==self.FOURSEASONSKEY:
            difsns = difsns[self.FOURSEASONS_COLNAMES]
        return difsns

    @mpl.rc_context(plt.style.use(MPL_STYLE_DEFAULT))
    def plot_time(self,figpath=None,figsize=None,colors=None):
        """Plot al heads in one graph and all head differences below that 
           in one figure 
           
        Parameters
        ----------
        figpath : str
            Valid filepath for output figure.
        figsize : [width,height]
            Figure size in inches.
        colors : list of tuples or strings
            Colors for head lines (RGB tuples or Hexcodes).

        Returns
        -------
        axes
        """

        headsdif = self.get_difference()
        seasons = self.get_seasons(dates=headsdif,period='biannual')

        # colors for heads lines
        if colors is None:
            cm = mpl.colormaps['tab10_r'].reversed() #mpl colormap
            colorlist = cm.colors #list of RGB-tuples
            
            colors = {}
            for i,col in enumerate(list(headsdif)):
                colors[col] = colorlist[i]

        # colors for seasonal colored differences
        """
        cm2 = mpl.colormaps['tab10_r']
        season_colors ={
            'Winter':cm2.colors[0],
            'Summer':cm2.colors[2],
            }
        """
        season_colors = self.TWOSEASONS_COLORS

        # figure settings
        nrows = len(list(self.heads))
        fig,axs = plt.subplots(nrows=nrows,ncols=1,sharex=True,constrained_layout=True)
        if figsize is None:
            figsize = [7,5]
        fig.set_figwidth(figsize[0])
        fig.set_figheight(figsize[1])
        ##plt.subplots_adjust(hspace=0.4)

        # plot heads on first ax
        for col in list(self.heads):
            heads = self.heads[col].dropna().squeeze()
            #heads.plot(ax=axs[0],color=colors[col])
            axs[0].plot(heads.index,heads,label=col,color=colors[col])
        axs[0].set_title(label=f'Location {self.locname}',fontsize=10.)
        axs[0].yaxis.set_label_text('stijghoogte', fontsize=7.)
        ncols = min(len(list(self.heads)),4)
        legend = axs[0].legend(loc='upper left', bbox_to_anchor=(0.0, 0.0),
            fancybox=False, shadow=False, edgecolor="white",
            ncols=ncols,fontsize=7.)
        legend.get_frame().set_alpha(None)
        legend.get_frame().set_facecolor((0, 0, 0, 0.0))

        # plot head differences on next axes
        colnames = [x for x in list(self.heads) if x!=self.refcol]
        for i,col in enumerate(colnames):

            sr = headsdif[col].dropna()*100
            periods = self.get_seasons(dates=sr,period='biannual')
            srw = sr[periods=='Winter']
            srs = sr[periods=='Summer']

            srw.plot(ax=axs[i+1],color=season_colors['Winter'],style='.',ms=2,label='winter')
            srs.plot(ax=axs[i+1],color=season_colors['Summer'],style='.',ms=2,label='zomer')

            axs[i+1].axhline(y=0, color='darkgray',linestyle='--')
            titletext = f'stijghoogteverschil {sr.name} met {self.refcol}'
            axs[i+1].set_title(titletext, fontsize=10.)
            axs[i+1].yaxis.set_label_text('stijghoogteverschil (cm)', fontsize=7.)

            axs[i+1].legend(['winter', 'zomer',], loc='lower left',ncols=2,edgecolor=None)

        # xax formatter
        half_year_locator = mdates.MonthLocator(bymonth=(4, 10))
        year_month_formatter = mdates.DateFormatter("%m\n%Y")
        for i in range(len(axs)):

            axs[i].set_xlabel(' ')
            axs[i].tick_params(axis='y', labelsize=7.)
            axs[i].tick_params(axis='x', which='both', bottom=False, labelsize=9.)

            if i==len(axs)-1: #last axis only
                axs[i].xaxis.set_major_locator(half_year_locator) # Locator for major axis only.
                axs[i].xaxis.set_major_formatter(year_month_formatter) # formatter for major axis only
                axs[i].xaxis.set_minor_locator(mdates.MonthLocator())
        
        for i in range(len(axs)):

                for label in axs[i].get_xticklabels(which='major'):
                    label.set(rotation=0, horizontalalignment='center')

        if figpath is not None:
            fig.savefig(figpath, dpi=300, facecolor='w', edgecolor='w')

        return axs

    @mpl.rc_context(plt.style.use(MPL_STYLE_DEFAULT))
    def plot_head(self,figsize=None,figpath=None,period='biannual'):
        """Plot head difference by reference head value.

        Parameters
        ----------
        figpath : str, optional

        period : {'seasonal,'biannual'}, default 'season'

        """
        headsdif = self.get_difference()
        relheads = self.get_relative_heads()
        colnames = headsdif.columns

        # create fig, ax
        if figsize is None:
            figsize=(8.5,8.5)

        if len(headsdif.columns) <= 2:
            ncols = 1
            nrows = 1
            figsize = figsize
        else:
            naxs = len(colnames)-1
            ncols = 2
            nrows = math.ceil(naxs/ncols)
            figsize = (figsize[0],figsize[1]/2*nrows)

        fig, axs = plt.subplots(
                    nrows, ncols, squeeze=False, 
                    sharex=True, sharey=False,
                    figsize=figsize, dpi=100)

        ax_idx = [(r,c) for r in range(nrows) for c in range(ncols)]

        if period=='biannual':
            #seasons = self.date_seasons(self._headsdif,periods='half-year')
            seasons = self.get_seasons(dates=headsdif,period='biannual')
            coldict = self.TWOSEASONS_COLORS
            colors = [coldict[x] for x in seasons]

        if period=='season':
            #seasons = self.date_seasons(self._headsdif,period='quarter')
            seasons = self.get_seasons(dates=headsdif,period='biannual')
            coldict = self.FOURSEASONS_COLORS
            colors = [coldict[x] for x in seasons]

        pal = sns.color_palette(colors)

        for i in range(len(colnames)-1):

            ax = axs[ax_idx[i]]

            ax.axhline(y=0, color='darkgray',linestyle='--')
            ax.axvline(x=0, color='darkgray',linestyle='--')

            ax.axhline(y=5, color='slategray',linestyle='-.')
            ax.axhline(y=-5, color='slategray',linestyle='-.')

            data = headsdif
            data[self.refcol] = relheads[self.refcol]

            ax = sns.scatterplot(
                    x = colnames[0], 
                    y = colnames[i+1], 
                    data=data*100, 
                    ax=ax,
                    #color=colors[i+1])
                    palette=coldict,
                    hue = seasons)

            handles, labels = ax.get_legend_handles_labels()
            # sort both labels and handles by labels
            labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
            ax.legend(handles, labels)

            ax.set_title(colnames[i+1])
            xlab = f'relatieve stijghoogte {colnames[0]} (cm)'
            ax.xaxis.set_label_text(xlab)
            ylab = 'stijghoogteverschil (cm)'
            ax.yaxis.set_label_text(ylab)

        fig.tight_layout()
        if figpath:
            fig.savefig(figpath, dpi=300)

        return axs

    @mpl.rc_context(plt.style.use(MPL_STYLE_DEFAULT))
    def plot_freq(self):
        """
        Plot head differences as grid of frequency plots.

        Returns
        -------
        fig,ax
        """
        headsdif = self.get_difference()
        ##ref = self.ref

        loclist = list(headsdif.columns)
        nrows = len(loclist)
        ncols = len(loclist)
        fig, axs = plt.subplots(nrows=nrows,ncols=ncols,sharex=True)

        # space between subplots
        plt.tight_layout()
        bbox = axs[0,0].get_position()
        hpad = bbox.width/100.*10.
        vpad = bbox.height/100.*10.
        plt.subplots_adjust(bottom=0.1,top=0.9,left=0.1,wspace=vpad, 
                            hspace=hpad)

        for i,ref in enumerate(loclist):

            tbl = headsdif  #(ref)
            for j,col in enumerate(loclist):
                x = (tbl[col]-tbl[col].mean()).values*100
                
                xmax=22.5
                dx = 5.

                nbars = math.ceil((xmax-dx/2)/dx)
                xup = nbars*dx+dx/2
                xlow = -xmax
                
                bin_list = np.arange(xlow,xup+1,dx)
                n, bins, patches = axs[i,j].hist(x, bin_list, density=1,)

                # color of rectangles (this is a slow solution)
                for k,pt in enumerate(patches):
                    if k < len(bin_list)/2-1:
                        pt.set_facecolor('#4ca3dd')
                    elif k == len(bin_list)/2-1:
                        pt.set_facecolor('#ccdee9') # #c0d6e4
                    else:
                        pt.set_facecolor('#ff7f50')

                axs[i,j].set_yticks([])
                axs[i,j].set_xticks([])

                if i==0:
                    axs[i,j].set_title(f'{col}',fontsize=18)
                if j==0:
                    axs[i,j].set_ylabel(f'{ref}',fontsize=18)

                # title below graphs
                if i ==nrows-1:
                    bbox = axs[i,j].get_position()
                    vpad = bbox.height/10.
                    mid = bbox.xmin+bbox.width/2.
                    plt.figtext(mid,bbox.ymin-vpad,f'{dx} cm',
                                fontsize='large',
                                horizontalalignment='center')

        return axs



""" 
This module containes the class GwSeries for managing a head series.

Examples
--------
gw = GwSeries.from_dinogws(<filepath to dinocsv file>)
gw = GwSeries.from_json(<filepath to acequia json file>)
""" 

import math
import os
import os.path
from collections import OrderedDict
import json
import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt

from pandas import Series, DataFrame
import pandas as pd
import numpy as np

##from .read import dinogws
from .plots import plotheads as plotheadsmodule
from .stats.gxg import GxgStats
from .stats.gwtimestats import GwTimeStats


class GwSeries:
    """ 
    Groundwater heads time series management

    Methods
    -------
    from_dinogws(filepath)
        read heads series from dinoloket csv file

    from_json(filepath)
        read heads series from json file

    to_json(filepath)
        read heads series from json file

    to_csv(filepath)
        read heads series from json file

    heads(ref,freq)
        return timeseries with measured heads

    name()
        return heads series name

    locprops(minimal)
        return location properties, optional minimal=True

    tubeprops(last)
        return tube properties, optinal only last row (last=True)

    stats(ref)
        return descriptice statistics

    describe()
        return selection of properties and descriptive statistics

    gxg()
        return tabel with gxg (desciptive statistics for groundwater 
        series used in the Netherlands)

    Examples
    --------
    To create a GwSeries object from file:  
    >>>gw = GwSeries.from_dinogws(<filepath to dinocsv file>)  
    >>>gw = GwSeries.from_json(<filepath to acequia json file>)  

    To get GwSeries properties:  
    >>>GwSeries.heads()  
    >>>GwSeries.locprops()  
    >>>GwSeries.name()  
    >>>GwSeries.heads1428()  

    To export GwSeries data:  
    >>>GwSeries.to_csv(<filename>)  
    >>>GwSeries.To_json(<filename>)  

    Notes
    -----
    Head measurements are stored in meters relative to welltop 
    and served in several units: mwelltop,mref,msurfacelevel.

    Valid row names for locprops and column names for tubeprops are
    stored in class variables locprops_names and tubeprops_names:
    >>> print(acequia.GwSeries.locprops_names)
    >>> print(acequia.GwSeries.tubeprops_names)
    """
    _headprops_names = [
        'headdatetime','headmp','headnote','remarks'
        ]

    _locprops_names = [
        'locname','filname','alias','xcr','ycr','height_datum',
        'grid_reference'
        ]

    _locprops_minimal = [
        'locname','filname','alias','xcr','ycr'
        ]

    _tubeprops_names = [
        'startdate','mplevel','filtop','filbot','surfacedate',
        'surfacelevel'
        ]

    _tubeprops_minimal = [
        'mplevel','surfacelevel','filbot',
        ]

    _tubeprops_numcols = [
        'mplevel','surfacelevel','filtop','filbot'
        ]

    _reflevels = [
        'datum','surface','mp',
        ]


    def __init__(self,heads=None,locprops=None,tubeprops=None):
        """
        Parameters
        ----------
        heads : pandas.DataFrame
            timeseries with groundwater heads
        locprops : pandas.Series
            series with location properties
        tubprops : pandas.DataFrame
            dataframe with tube properties in time

        """

        if locprops is None:
            self._locprops = Series(index=self._locprops_names,
                dtype='object')
        elif isinstance(locprops,pd.Series):
            self._locprops = locprops
        else:
            raise TypeError(f'locprops is not a pandas Series but '
                f'{type(locprops)}')

        if tubeprops is None:
            self._tubeprops = DataFrame(columns=self._tubeprops_names)
        elif isinstance(tubeprops,pd.DataFrame):
            self._tubeprops = tubeprops
        else:
            raise TypeError(f'tubeprops is not a pandas DataFrame '
                f'but {type(tubeprops)}')

        if heads is None: 
            self._heads = pd.DataFrame(columns=self._headprops_names) #Series()
            self._heads_original = self._heads.copy()

        elif isinstance(heads,pd.DataFrame):
            self._heads = heads
            self._heads_original = self._heads.copy()

        else:
            raise TypeError(f'heads is not a pandas DataFrame but {type(heads)}')

    def __repr__(self):
        return (f'{self.name()} (n={len(self)})')

    def __len__(self):
        return len(self._heads)

    def _validate_reference(self,ref):

        if ref is None:
            return self._reflevels[0]

        if ref not in self._reflevels:
            warnings.warn((f'Reference level {ref} is not valid.'
                f'Reference level {self._reflevels[0]} is assumed.'),
                stacklevel=2)
            return self._reflevels[0]

        return ref

    @classmethod
    def from_json(cls,filepath=None):
        """ Read gwseries object from json file """

        with open(filepath) as json_file:
            json_dict = json.load(json_file)

        locprops = DataFrame.from_dict(json_dict['locprops'],
                                        orient='index')
        locprops = Series(data=locprops[0],index=locprops.index,
                                        name='locprops')

        tubeprops = DataFrame.from_dict(json_dict['tubeprops'],
                    orient='index')
        tubeprops.name = 'tubeprops'
        tubeprops['startdate'] = pd.to_datetime(tubeprops['startdate']) #.dt.date

        heads = DataFrame.from_dict(json_dict['heads'],orient='index')

        return cls(heads=heads,locprops=locprops,tubeprops=tubeprops)


    def to_json(self,filepath=None):
        """ 
        Create json string from GwWeries object and optionally write 
        to file.

        Parameters
        ----------
        path : str
           Filepath or directory the json file will be written to.
           If filepath is not given no textfile will be written and 
           only OrderedDict with valid JSON wil be retruned.

        Returns
        -------
        OrderedDict with valid json

        Notes
        -----
        If no value for dirpath is given, a valid json string is
        returned. If a value for dirpath is given, nothing is returned 
        and a json file will be written to a file with the series name
        in dirpath.
        """
        json_locprops = json.loads(
            self._locprops.to_json()
            )
        json_tubeprops = json.loads(
            self._tubeprops.to_json(date_format='iso',orient='index')
            )
        json_heads = json.loads(
            self._heads.to_json(date_format='iso',orient='index',
            date_unit='s')
            )

        json_dict = OrderedDict()
        json_dict['locprops'] = json_locprops
        json_dict['tubeprops'] = json_tubeprops
        json_dict['heads'] = json_heads
        json_formatted_str = json.dumps(json_dict, indent=2)

        if os.path.isdir(filepath):
            filepath = os.path.join(filepath,self.name()+'.json')

        try:            
            with open(filepath,"w") as f:
                f.write(json_formatted_str)
        except FileNotFoundError:
            print("Filepath {} does not exist".format(filepath))
            return None
        #finally:
        #    json_dict

        return json_dict


    def to_csv(self,path=None,ref=None):
        """
        Export groundwater heads series to simple csv file.

        Parameters
        ----------
        path : str
            csv file wil be exported to path, if path is a directory,
            series will be saved as <path><name>.csv.
            if path is not given, file is saved in present directory.

        Examples
        --------
        Save heads to simple csv:

        >>>aq.GwSeries.to_csv(<dirpath>)

        Read back with standard Pandas:

        >>>pd.read_csv(<filepath>,  parse_dates=['date'], 
                  index_col='date', squeeze=True) 
        
        """
        self._csvpath = path
        self._csvref = ref

        if self._csvpath is None:
            self._csvpath = f'{self.name()}.csv'

        if os.path.isdir(self._csvpath):
            filename = f'{self.name()}.csv'
            self._csvpath = os.path.join(self._csvpath,filename)

        try:
            sr = self.heads(ref=self._csvref)
            sr.to_csv(self._csvpath,index=True,index_label='datetime',
                header=['head'])
        except FileNotFoundError:
            msg = f'Filepath {self._csvpath} not found'
            warnings.warn(msg)
            result = None
        else:
            result = sr

        return result


    def name(self):
        """ Return groundwater series name """
        location = str(self._locprops['locname'])
        filter = str(self._locprops['filname'])
        return location+'_'+filter


    def locname(self):
        """Return series location name"""
        ##srname = self.locprops().index[0]
        ##locname = self.locprops().loc[srname,'locname']
        return self._locprops['locname']


    def locprops(self):
        """Return location properties as pd.DataFrame."""
        locprops = self._locprops
        locprops = locprops.drop('filname')
        locprops = DataFrame(locprops).T.set_index('locname')
        return locprops


    def tubeprops(self,last=False,minimal=False):
        """
        Return tube properties.

        Parameters
        ----------
        last : booean, default False
            retun only the last row of tube properties without date

        minimal : bool, default False
            return only minimal selection of columns

        Returns
        -------
        pd.DataFrame
        """
        tps = DataFrame(self._tubeprops[self._tubeprops_names]).copy()
        #tps['startdate'] = tps['startdate'].dt.date
        tps['startdate'].apply(pd.to_datetime, errors='coerce')

        if minimal:
            tps = tps[self._tubeprops_minimal]

        tps.insert(0,'series',self.name())

        if last:
            #tps = tps.iloc[[-1]]
            tps = tps.tail(1)
            #tps = tps.set_index('series')

        return tps

    def surface(self):
        """Return last known surface level"""
        return self._tubeprops['surfacelevel'].iat[-1]


    def heads(self,ref='datum',freq=None):
        """ 
        Return groundwater head measurements.

        Parameters
        ----------
        ref  : {'mp','datum','surface'}, default 'datum'
               choosen reference for groundwater heads
        freq : None or any valid Pandas Offset Alias
                determine frequency of time series

        Returns
        -------
        result : pandas time Series

        Notes
        ----=
        Parameter 'ref' determines the reference level for the heads:
        'mp'   : elative to well top ('measurement point')
        'datum': relative to chosen level (would be meter +NAP for the
                 Netherlands, or TAW for Belgium)
        'surface' : relative to surface level (meter min maaiveld)

        Parameter 'freq' determines the time series frequency by setting
        the Pandas Offset Alias. When 'freq' is None, no resampling is
        applied.
        Valid values for 'freq' would be:
        'H' : hourly frequency
        'D' : calender day frequency
        'W' : weekly frequency
        'M' : month end frequency
        'MS': month start freuency
        'Q' : quarter end frequency
        'QS': quarter start frequency
        'A' : year end frequency
        'AS': year start frequency
        """
        if not ref:
            ref = 'datum'

        if ref not in self._reflevels:
            msg = f'{ref} is not a valid reference level name'
            raise ValueError(msg)

        heads = self._heads[['headdatetime','headmp']]
        heads = heads.set_index('headdatetime',drop=True) ##.squeeze()
        heads.name = self.name()

        headscopy = heads.copy()

        if ref in ['datum','surface']:

            srvals = headscopy.values.flatten()
            srvals2 = headscopy.values.flatten()
            for index,props in self._tubeprops.iterrows():

                mask = heads.index>=props['startdate']
                if ref=='datum':

                    if not pd.api.types.is_number(props['mplevel']):
                        msg = f'{self.name()} tubeprops mplevel is None.'
                        warnings.warn(msg)
                        mp = 0
                    else:
                        mp = props['mplevel']

                    srvals2 = np.where(mask,mp-srvals,srvals2)


                if ref=='surface':
                    if not pd.isnull(props['surfacelevel']):
                        surfref = round(props['mplevel']-props['surfacelevel'],2)
                        srvals2 = np.where(mask,srvals-surfref,srvals2)

                    else:
                        msg = f'{self.name()} surface level is None'
                        warnings.warn(msg)
                        srvals2 = np.where(mask,srvals,srvals)

            heads = Series(srvals2,index=heads.index)
            heads.name = self.name()

        if freq is not None:
            heads = heads.resample(freq).mean()
            heads.index = heads.index.tz_localize(None)

        return heads


    def timestats(self,ref=None):
        """
        Return descriptive statistics for heads time series.

        Parameters
        ----------
        ref  : {'mp','datum','surface'}, default 'datum'
            reference level for groundwater heads

        Returns
        -------
        pd.Series
        """

        self._ref = self._validate_reference(ref)

        ts = self.heads(ref=ref)
        gwstats = GwTimeStats(ts)

        return gwstats.stats()


    def describe(self,ref='datum',gxg=False,minimal=True):
        """
        Return selection of properties and descriptive statistics.

        Parameters
        ----------
        ref  : {'mp','datum','surface'}, default 'datum'
            choosen reference level for groundwater heads
        gxg : bool, default False
            add GxG descriptive statistics
        minimal : bool, default True
            return minimal selection of statistics

        Returns
        -------
        pd.Series
        """

        self._ref = self._validate_reference(ref)

        srlist = []

        srlist.append(self._locprops[self._locprops_minimal])

        tubeprops = (self._tubeprops[self._tubeprops_minimal].tail(1
            ).iloc[0,:])
        srlist.append(tubeprops)

        timestats = self.timestats(ref=self._ref)
        srlist.append(timestats)

        if gxg==True:
            gxg = self.gxg(ref=self._ref,minimal=minimal)
            srlist.append(gxg)

        sr = pd.concat(srlist,axis=0)
        sr.name = self.name()

        if self._ref=='surface':

            for key in ['filbot']:
                sr[key] = (sr['surfacelevel']-sr[key])*100

            for key in ['mean','median','q05','q95','dq0595']:
                sr[key] = sr[key]*100

            for key in ['filbot','mean','median','q05','q95',
                'dq0595','n1428']:
                if not np.isnan(sr[key]):
                    sr[key] = math.floor(sr[key])

        """
        locprops = self.locprops(minimal=minimal)
        tubeprops = self.tubeprops(last=True,minimal=True)
        tubeprops = tubeprops.set_index('series')

        tbl = pd.merge(locprops,tubeprops,left_index=True,right_index=True,how='outer')

        srstats = self.timestats(ref=ref)
        tbl = pd.merge(tbl,srstats,left_index=True,right_index=True,how='outer')

        if gxg==True:
            gxg = self.gxg()
            tbl = pd.merge(tbl,gxg,left_index=True,right_index=True,how='left')
        """
        return sr


    def tubeprops_changes(self,proptype='mplevel',relative=True):
        """
        Return timeseries with tubeprops changes.

        Parameters
        ----------
        proptype : ['mplevel','surfacelevel','filtop','filbot'
            Tubeproperty that is shown in reference cange graph.
        relative : bool, default True
            Changes relative to first value.

        Returns
        -------
        pd.Series
        """
        if proptype in ['mplevel','surfacelevel','filtop','filbot']:
            mps = self._tubeprops[proptype].values
        else:
            mps = self._tubeprops['mplevel']
            # TODO: add userwarning

        idx = pd.to_datetime(self._tubeprops['startdate'])
        sr1 = Series(mps,index=idx)

        idx = sr1.index[1:]-pd.Timedelta(days=1)
        lastdate = self.heads().index[-1]
        idx = idx.append(pd.to_datetime([lastdate]))
        sr2 = Series(mps,index=idx)

        sr12 = pd.concat([sr1,sr2]).sort_index()
        if relative:
            sr12 = sr12 - sr12[0]

        return sr12


    def plotheads(self,proptype=None,filename=None):
        """
        Plot groundwater heads time series.

        Parameters
        ----------
        proptype : ['mplevel','surfacelevel','filtop','filbot'
            Tubeproperty that is shown in reference cange graph.
            If not given, no reference plot will be shown.
        """
        if proptype in ['mplevel','surfacelevel','filtop','filbot']:
            ##mps = self._tubeprops[proptype].values
            mps = self.tubeprops_changes(proptype=proptype)
            self.headsplot = plotheadsmodule.PlotHeads(ts=[self.heads()],mps=mps)

        if proptype is None:
            self.headsplot = plotheadsmodule.PlotHeads(ts=[self.heads()])

        if filename is not None:
            self.headsplot.save(filename)

        return self.headsplot


    def gxg(self,ref='datum',minimal=True,name=True):
        """
        Return tables with desciptive statistics GxG and xG.

        Parameters
        ----------
        ref : {'datum','surface'}, default 'datum'
            Reference level for gxg statistics.
        minimal : bool, default False
            Return minimal set of statistics.
        name : bool, default True
            Include series name in multiindex of xg.

        Returns
        -------
        gxg : pd.Series
            gxg descriptive statistics
        """
        # Calculating gxg can take considerable time. Therefore,
        # results are stored.
        if not hasattr(self,'_gxg'):
            self._gxg = GxgStats(self)            

        gxg = self._gxg.gxg(reference=ref,minimal=minimal)
        
        return gxg


    def xg(self,ref='datum',name=True):
        """
        Return tables with xg desciptive statistics for each year.

        Parameters
        ----------
        ref : {'datum','surface'}, default 'datum'
            Reference level for gxg statistics.
        name : bool, default True
            Include series name in multiindex of xg.

        Returns
        -------
        xg : pd.DataFrame
        """
        if not hasattr(self,'_gxg'):
            self._gxg = GxgStats(self)            

        xg = self._gxg.xg(reference=ref,name=name)

        return xg

    def get_quantiles(self, ref='surface', unit='days', step=None):
        """Calculate quantiles of measured heads.

        Parameters
        ----------
        headsref : {'datum','surface'}, default 'surface'
            Reference level for measurements.
        unit : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.

        Returns
        -------
        pandas.DataFrame
        """
        # bypass circular import
        from .stats.quantiles import Quantiles

        qt = Quantiles(self.heads(ref=ref))
        return qt.get_quantiles()

    def get_ecostats(self, ref='surface', units='days', step=5):
        """Return ecological most relevant statistics.

        Parameters
        ----------
        ref : {'datum','surface'}, default 'surface'
            Reference level for measurements.
        units : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int, default 5
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.

        Returns
        -------
        pandas.Series      
        """

        # bypass circular import
        from .stats.quantiles import Quantiles
        
        qt = Quantiles(self.heads(ref=ref))
        inundation = qt.get_inundation()
        lowest = qt.get_lowest()

        ecostats = self.gxg(ref='surface',minimal=True)
       
        ecostats['lowest_mean'] = lowest['mean']
        ecostats['lowest_min'] = lowest['min']
        ecostats['lowest_max'] = lowest['max']

        ecostats['inundation_mean'] = inundation['mean']
        ecostats['inundation_min'] = inundation['min']            
        ecostats['inundation_max'] = inundation['max']

        return ecostats



import numpy as np
from pandas import DataFrame, Series
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from acequia import GwSeries

def plot_tubechanges(gw=None,ax=None):
    """Time series plot of heads and tube changes.
    
    Parameters
    ----------
    gw : GwSeries
        Groundwater heads and tube properties.
    ax : 

    Returns
    -------
    ax
    """
    headsclr = '#3e92d1'
    welltopclr = '#f44336'
    srfclr = 'gray'
    filbotclr = '#6a329f'

    #if not isinstance(gw,GwSeries):

    heads = gw.heads(ref='datum')
    tubeprops = gw.tubeprops()
    mp = gw.tubeprops_changes(proptype='mplevel',relative=False)
    sf = gw.tubeprops_changes(proptype='surfacelevel',relative=False)
    fb = gw.tubeprops_changes(proptype='filbot',relative=False)

    heads.plot(ax=ax,color=headsclr,label='stijghoogte')
    mp.plot(ax=ax,color=welltopclr,label='bovenkant buis',)
    fb.plot(ax=ax,color=filbotclr,label='onderkant filter')
    sf.plot(ax=ax,color=srfclr,label='maaiveld',linestyle='--')

    plt.setp(ax.get_xticklabels(), rotation=0, ha='center')
    ax.set_xlabel('')

    ax.xaxis.set_major_locator(mdates.YearLocator(3))
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y'))

    ax.legend(loc='upper center',bbox_to_anchor=(0.5, -0.15), ncol=4,
        prop={'size':9.})
    ax.set_title(gw.name(),fontdict={'fontsize':11.},loc='left')

    mpdate = fb.index[0]

    for idx,row in gw.tubeprops().iterrows():
        plt.axvline(x=row['startdate'],color="gray", linestyle=":")

        dx = pd.Timedelta(days=250)
        x = row['startdate']-dx
        dy = (row['mplevel']-row['filbot'])*0.05
        y = row['filbot']+dy
        text = row['startdate'].strftime('%d-%m-%Y')
        plt.text(x,y,text,rotation='vertical',fontsize='small')

    x = heads.index[-1]
    text = heads.index[-1].strftime('%d-%m-%Y')
    plt.axvline(x=x,color="gray", linestyle=":")
    plt.text(x-dx,y,text,rotation='vertical',fontsize='small')
    
    return ax




#print(f'Loading package {__name__}')

import logging

from .gwseries import GwSeries
from .gwcollection import GwCollection
from .gwlist import GwList
from .gwlist import headsfiles as headsfiles
from .gwlocs import GwLocs
from .read.gwfiles import GwFiles
from .headsdif import HeadsDif
from .swseries import SwSeries
from .data.knmidata import KnmiData
from .geo.coordinate_conversion import CrdCon
from .geo.waypoint_kml import WpKml
from .geo.pointshapewriter import PointShapeWriter, pointshape_write
from .plots.plotheads import PlotHeads
from .plots.tsmodelstatsplot import TsModelStatsPlot,plot_tsmodel_statistics
from .plots.plotfun import plot_tubechanges
from .read.dinogws import DinoGws
from .read.dinosurfacelevel import DinoSurfaceLevel
from .read.dawaco import Dawaco
from .read.gpxtree import GpxTree
from .read.hydromonitor import HydroMonitor
from .read.waterweb import WaterWeb
from .read.waterwebtools import measurement_types
from .read.knmi_weather import KnmiWeather
from .read.knmi_rain import KnmiRain
from .read.knmi_download import KnmiDownload
from .read.filedirtools import listdir, cleardir
from .read.brogldxml import BroGldXml
from .read.brogmwxml import BroGmwXml
from .read import brorest
from .stats.utils import hydroyear, season, index1428, ts1428
from .stats.utils import measfrq, maxfrq
from .stats.gwtimestats import GwTimeStats, gwtimestats
from .stats.gxg import GxgStats, stats_gxg
from .stats.gwliststats import GwListStats, gwliststats, gwlocstats
from .stats.quantiles import Quantiles
from .stats.meteo_drought import MeteoDrought

__all__ = ['GwSeries','GwList','GwLocs','PlotGws','DinoGws', 
           'SwSeries','DinoSurfaceLevel','WaterWebNetwork',
           'HydroMonitor','TimeStats','PointShapeWriter','GxgStats',
           'Quantiles','CrdCon','BroGldXml',
           'KnmiWeather','KnmiRain', 'KnmiStations',
           'GwListStats','gwliststats','gwlocstats',
           'pointshape_write',
           'TsModelStatsPlot','plot_tsmodel_statistics',
           'MeteoDrought','GpxTree']

logger = logging.getLogger(__name__)



""" Read groundwater measurement data from dinoloket csv files.

  Author : Thomas de Meij, 2020
  
  History: 02-02-2014 created for python2.7;
           15-08-2015 migrated to python3.x;     
           06-07-2019 migrated to acequia 

"""

import os
import os.path
from datetime import datetime
from collections import OrderedDict
import csv
import time
import datetime as dt
import warnings
import numpy as np
from pandas import Series, DataFrame
import pandas as pd

from ..gwseries import GwSeries

sep = ","

class DinoGws:
    """Read TNO Dinoloket csv file with groundwater measurement data"""

    METATAG = ','.join(
        ['Locatie','Filternummer','Externe aanduiding',
         'X-coordinaat','Y-coordinaat','Maaiveld (cm t.o.v. NAP)',
         'Datum maaiveld gemeten','Startdatum','Einddatum',
         'Meetpunt (cm t.o.v. NAP)','Meetpunt (cm t.o.v. MV)',
         'Bovenkant filter (cm t.o.v. NAP)',
         'Onderkant filter (cm t.o.v. NAP)'
        ])

    DATATAG = ','.join(
        ['Locatie','Filternummer','Peildatum',
         'Stand (cm t.o.v. MP)','Stand (cm t.o.v. MV)',
         'Stand (cm t.o.v. NAP)','Bijzonderheid,Opmerking','','',''
         ])

    MISSINGDATA = (f'Van deze put zijn geen standen opgenomen',
         f'in de DINO-database')

    FILTERCOLS = ["nitgcode","filter","tnocode","xcoor",
        "ycoor","mvcmnap","mvdatum","startdatum","einddatum",
        "mpcmnap","mpcmmv","filtopcmnap","filbotcmnap"]

    DATACOLS = ["nitgcode","filter","peildatum","standcmmp",
        "standcmmv","standcmnap","bijzonderheid","opmerking"]

    HEADCOLS = ['peildatum','standcmmp','bijzonderheid','opmerking']


    MAPPING_DINOHEADPROPS = OrderedDict([
        ("headdatetime","peildatum"),("headmp","standcmmp"),
        ("headnote","bijzonderheid"),("remarks","opmerking"),
        ])

    MAPPING_DINOLOCPROPS = OrderedDict([
        ('locname','nitgcode'),
        ('filname','filter'),
        ('alias','tnocode'),
        ('xcr','xcoor'),
        ('ycr','ycoor'),
        ('height_datum','NAP'),
        ('grid_reference','RD'),
        ])

    MAPPING_DINOTUBEPROPS = OrderedDict([
        ('startdate','startdatum'),
        ('mplevel','mpcmnap'),
        ('filtop','filtopcmnap'),
        ('filbot','filbotcmnap'),
        ('surfacedate','mvdatum'),
        ('surfacelevel','mvcmnap'),
        ])

    def __init__(self,filepath=None,readall=True):
        """
        Parameters
        ----------
        filepath : str
            Valid filepath to dinoloket csv file.
        readall : bool, default True
            Read all data (True) or header only (False).
        ...
        """
        self.filepath = filepath

        if isinstance(readall,bool):
            self.readall=readall
        else:
            warnings.warn((f"Variable {readall} not of type 'bool' "
                     "but type '{readall}'. Data wil be read."))

        # create empty variables
        self.errors = []
        self._data = DataFrame()
        self._header = DataFrame()       
        self._datatext = DataFrame()
        self._headertext = DataFrame()
        self.dfdesc = DataFrame()
        self.dfdescloc = DataFrame()
        self.seriesname = ""
        
        if filepath != None:
            self.flines = self._readfile(filepath=self.filepath)
            self._header, self._data = self._readlines()

        if self._header.empty and not self._data.empty:
            self._header = DataFrame(data=[[np.nan]*len(self.FILTERCOLS)],columns=self.FILTERCOLS)
            self._header = self._header.astype({'nitgcode':str,'filter':str})
            self._header.at[0,'nitgcode'] = self._data.at[0,'nitgcode']
            self._header.at[0,'filter'] = self._data.at[0,'filter']
            self._header.at[0,'startdatum'] = self._data.at[0,'peildatum']

    def __repr__(self):
        return (f'{self.srname} (n={len(self._data)})')

    @property
    def gwseries(self):
        """Return GwSeries object."""

        # get location metadata
        locprops = Series(index=GwSeries._locprops_names,dtype='object')

        for propname in GwSeries._locprops_names:
            dinoprop = self.MAPPING_DINOLOCPROPS[propname]
            if dinoprop in self.FILTERCOLS:
                locprops[propname] = self.header.at[0,dinoprop]

        locprops['grid_reference'] = 'RD'
        locprops['height_datum'] = 'mNAP'
        locprops = Series(locprops)

        # get piezometer metadata
        tubeprops = DataFrame(columns=GwSeries._tubeprops_names)
        for prop in GwSeries._tubeprops_names:
            dinoprop = self.MAPPING_DINOTUBEPROPS[prop]
            if dinoprop in self.FILTERCOLS:
                tubeprops[prop] = self.header[dinoprop]

        for col in GwSeries._tubeprops_numcols:
                tubeprops[col] = pd.to_numeric(tubeprops[col],
                                 errors='coerce')/100.

        # get head measurements
        heads = DataFrame(columns=GwSeries._headprops_names)
        for prop in GwSeries._headprops_names:
            dinoprop = self.MAPPING_DINOHEADPROPS[prop]
            if dinoprop in self.HEADCOLS:
                heads[prop] = self.headdata[dinoprop]
        heads['headmp'] = heads['headmp']/100.

        return GwSeries(heads=heads, locprops=locprops, tubeprops=tubeprops)


    def _readfile(self,filepath=None):
        """ Open DINO csv file and return list of filelines """

        try:
            file = open(filepath,'r')
        except (IOError, TypeError) as err:
            errno, strerror = err.args
            print("{!s}".format(errno), end="")
            print("I/O fout{!s}".format(strerror), end="")
            print (" : "+filepath)
            self.errors.append(
                    [filepath,
                     "File can not be opened"])
            self.flines=[]
            raise
        else:
            self.flines = file.readlines()
            file.close()

        return self.flines


    def _readlines(self):
        """ read list of file lines from dinofile to data """

        # assert file is valid ascii
        if len(self.flines)==0:
            file_valid = False
        elif self.flines[0][0]=='\x00':
            # test for corrupted file with only 'x00' 
            # Yes, I have really seen this
            file_valid=False
        else:
            file_valid = True

        if file_valid==False:
            self.headerstart=0
            self.headerend=0
            self.datastart=0
        else:
            # findlines in dinofile lines
            self.headerstart, self.headerend, self.datastart = \
                self._findlines()
        
        # read header
        if self.headerstart>0 and self.headerend>0: 
            self._header = self._readheader()
        else:
            self._header = DataFrame()

        # read data
        if self.datastart>0:
            self._data = self._readgws()
        else:
            self._data = DataFrame()

        return self._header, self._data


    def _findlines(self):
        """ Find start of header and data; if file has data at all """

        # set variables to find at zero
        self.headerstart = 0
        self.headerend = 0
        self.datastart = 0

        # find variables
        for il in range(len(self.flines)):

            if self.flines[il].startswith(self.MISSINGDATA):
                # put zonder gegevens
                self.errors.append([self.filepath,"Bestand bevat geen data"])
                self.hasheader = False
                self.hasdata = False
                break

            if self.flines[il].startswith(self.METATAG): #("Locatie,Filternummer,Externe"):
                if not self.flines[il+1].startswith("B"): # er zijn geen headerlines onder de headerkop
                    self.hasheader = False
                    self.errors.append([self.filepath,"Bestand zonder header"])                               
                else:
                    while True:
                        il+=1
                        if self.flines[il].startswith("B"):
                            if self.headerstart==0:
                                self.hasheader = True                        
                                self.headerstart = il
                        else: #voorbij laatste regel header
                            self.headerend = il
                            break
                            
            if self.flines[il].startswith(self.DATATAG):
                # bepaal eerste regelnummer met data
                il+=1
                if self.flines[il].startswith("B"):
                    self.hasdata = True
                    self.datastart = il
                else:
                    self.hasdata = False
                    self.errors.append([self.filepath,"Bestand zonder grondwaterstanden"])
                break
        il+=1
        # end of def findlines
        return self.headerstart, self.headerend, self.datastart

    @staticmethod
    def parse_dino_date(datestring,addtime=False):
        if isinstance(datestring, str):
            if datestring!="":                
                # string to datetime.datetime object
                if addtime==True: date = datetime.strptime(
                    datestring+" 12:00", "%d-%m-%Y %H:%M")
                else: date = datetime.strptime(datestring, "%d-%m-%Y")                    
            else:
                date = np.NaN
        else:
            date = np.NaN
        return date

    def _readheader(self): #public
        """ Read header data and return pandas dataframe """

        if self.headerstart>0 and self.headerend > self.headerstart:
            # create _header
            headerlist = [line[:-1].split(sep) for line in self.flines[self.headerstart:self.headerend]]
            self._header = DataFrame(headerlist, columns=self.FILTERCOLS)
            self._headertext = DataFrame(headerlist, columns=self.FILTERCOLS)

            # transform column values
            self._header["filter"] = self._header["filter"].apply(lambda x: x.lstrip("0"))
            self._header["mvdatum"] = self._header["mvdatum"].apply(lambda x:self.parse_dino_date(x))
            self._header["startdatum"] = self._header["startdatum"].apply(lambda x:self.parse_dino_date(x))
            self._header["einddatum"] = self._header["einddatum"].apply(lambda x:self.parse_dino_date(x))

        else:
            # create empty dataframe
            self._header = DataFrame(columns=self.FILTERCOLS)
            self._headertext = DataFrame(columns=self.FILTERCOLS)
            self.seriesname = "onbekend"

        return self._header


    def _readgws(self):
        """ Read groundwater measurements to pandas data frame """

        def fstr2float(astr):
            try:
                aval = float(astr)
            except ValueError:
                aval = np.NaN
            return aval


        if self.datastart>0:

            # create list of data from filelines
            data = [line[:-1].split(sep)[0:7]+[sep.join(line[:-1].split(
                sep)[7:])] for line in self.flines[self.datastart:]]            
            self._datatext = DataFrame(data, columns=self.DATACOLS)
            self._data = DataFrame(data, columns=self.DATACOLS)

            # transform column values
            self._data["peildatum"] = self._data["peildatum"].apply(
                lambda x:datetime.strptime(x,"%d-%m-%Y"))  # 28-03-1958
            self._data["filter"] = self._data["filter"].apply(
                lambda x: x.lstrip("0"))
            self._data["standcmmp"] = self._data["standcmmp"].apply(
                lambda x: fstr2float(x))
            self._data["standcmmv"] = self._data["standcmmv"].apply(
                lambda x: fstr2float(x))
            self._data["standcmnap"] = self._data["standcmnap"].apply(
                lambda x: fstr2float(x))
            self._data["opmerking"] = self._data["opmerking"].apply(
                lambda x: x.strip(","))

        else:
            self._data = DataFrame(columns=self.DATACOLS)
            self._datatext = DataFrame(columns=self.DATACOLS)
        return self._data

    def get_heads(self,units="cmmv"):
        """ Return time series with groundwater measurements.

        Parameters
        ----------
        units : {'cmmv','cmmp','cmnap'}, default 'cmmv'

        Returns
        -------
        pd.Series
        """
        # create series from _data
        if len(self._data)>0:
            if units=="cmmv":
                self.srseries = Series(self._data["standcmmv"].values, 
                    index=self._data["peildatum"],name=self.srname)
            if units=="cmmp":
                self.srseries = Series(self._data["standcmmp"].values, 
                    index=self._data["peildatum"],name=self.srname)
            if units=="cmnap":
                self.srseries = Series(self._data["standcmnap"].values, 
                    index=self._data["peildatum"],name=self.srname)
        else: # create empty series
            self.srseries = Series(name=self.srname)
        return self.srseries


    @property
    def headdata(self):
        """Return head data fromdino csv file"""
        if len(self._data)>0:
            data = self._data[self.HEADCOLS].copy()
        else:
            data = DataFrame(columns=self.HEADCOLS)
        return data

    @property
    def header(self):
        """ Return raw metadata from series as a pandas dataframe """        
        return self._header

    @property
    def data(self):
        """ Return raw data as Pandas dataframe """
        return self._data

    @property
    def locname(self): #, newname=None):
        ##if newname!=None: 
        ##    self.location = newname
        if not self.header.empty:
            self.location = self._header.loc[0,"nitgcode"]
        else:
            self.location = "B00A0000"
        return self.location

    @property
    def filname(self): #, newname=None):
        #if newname!=None: 
        #    self.filter = newname
        if not self.header.empty:
            self.filter = self.header.loc[0,"filter"]
        else: self.filter = "0"
        return self.filter

    @property
    def srname(self):
        self.seriesname = self.locname+"_"+self.filname
        return self.seriesname

    @property
    def describe(self):
        """ Return one line of metadata from series as a pandas dataframe """

        # deze functie moet eruit geschreven worden omdat parse_dino_date dit al afvangt
        def valdate(date):
            """ Validate datetime (no dates before 1900) """
            return (date.strftime("%d-%m-%Y") if date.year>1900 else np.NaN)

        if self.dfdesc.empty:
            filcols = ['reeksnaam', 'nitgcode', 'filter', 'tnocode', 'xcoor', 'ycoor','mvcmnap', 'mvdatum', 'startdatum', 'einddatum', 'mpcmnap', 'mpcmmv','filtopcmnap', 'filbotcmnap']
            self.dfdesc = DataFrame(columns=filcols)

            if len(self._readheader()) !=0 and len(self._readgws())!=0:

                # make one line of metadata
                dftail = self.header.tail(1).copy()

                # insert series name as first column
                ##dftail['reeksnaam'] = dftail["nitgcode"]+"_"+dftail["filter"].values.astype('str')
                dftail['reeksnaam'] = self.srname

                # recalculate startdate and enddate from measurements
                colnr = dftail.columns.get_loc("startdatum")
                startdatestr = self.data["peildatum"][0]
                dftail.iloc[0,colnr] = pd.to_datetime(valdate(startdatestr),format='%d-%m-%Y')
                
                colnr = dftail.columns.get_loc("einddatum")
                enddatestr   = self.data["peildatum"][len(self.data)-1]
                dftail.iloc[0,colnr]  = pd.to_datetime(valdate(enddatestr),format='%d-%m-%Y')

                self.dfdesc = dftail.reindex(columns=filcols).copy()
                self.dfdesc = self.dfdesc.reindex() # set index to 0

        return self.dfdesc

    @property
    def mpref(self):
        """ create dataframe with series of mp reference changes (for plotting line of ref changes above gwseries graph)"""
        msg = 'mpref method is depricates. use GwSeries.tubepropchanges() instead.'
        warnings.warn(msg, warnings) #.DeprecationWarning)


    def get_locations(self,df=DataFrame()):
        """ create table of locations from dataframe with data from several filters created by function describe() """

        warnings.warn('This method is deprecated. Use GwSeries instead.', 
            DeprecationWarning, stacklevel=2)
        return DataFrame()

        # code below is depricated
        # ------------------------

        # select one row for each group of filters
        grp = df.groupby("nitgcode")
        dfloc = grp.last()
        dfloc["startdatum"] = grp["startdatum"].min()
        dfloc["einddatum"] = grp["einddatum"].max()
        dfloc["nfil"] = pd.to_numeric(grp["filter"].count())

        # add index as column and reindex rows
        dfloc.insert(0,"nitgcode",dfloc.index)
        dfloc = dfloc.reset_index(drop=True)

        # drop columns with filter specific data
        dropcols = ["reeksnaam","filter",'mpcmnap','mpcmmv','filtopcmnap','filbotcmnap']
        dfloc = dfloc.drop(dropcols,axis=1,inplace=False) #,axis=1)

        # reorder columns
        dfloc = dfloc.reindex(columns=['nitgcode','tnocode','nfil','mvcmnap','mvdatum','startdatum','einddatum','xcoor','ycoor'])

        return dfloc


    def merge(self,dn2=None):
        """ add series of dinogws object to dinogws object and return merged object """

        warnings.warn('This method is deprecated. Use GwSeries instead.', 
            DeprecationWarning, stacklevel=2)
        return DinoGws()

        # code below is depricated
        # ------------------------

        if not isinstance(self,dinogws):
            raise TypeError("Input of dinogws.merge() must be another dinogws object")

        if (not self._data.empty) and (not dn2._data.empty):

            # determine possible overlap
            startdate1 = self.data().peildatum.min()
            enddate1 = self.data().peildatum.max()
            startdate2 = dn2.data().peildatum.min()
            enddate2 =  dn2.data().peildatum.max()
            if (startdate2 <= enddate1) and (startdate1 <= enddate2):
                raise ValueError('date ranges from both series show overlap')

            # merge headers
            self._header = pd.concat([self._header,dn2._header])
            self._header = self._header.sort_values(by=['startdatum']).reset_index()

            # merge groundwater measurement data
            self._data = pd.concat([self._data,dn2._data]).sort_values(by=['peildatum'])

        return self

def filesfromdir(dir):
    """Return list of dino sourcefiles from directory """

    filenames = []
    seriesnames = []
    for root, dirs, files in os.walk(dir):
        seriesnames += [f[0:11] for f in files if f[11:13]=="_1"]
        filenames += [os.path.join(root, f) for f in files if f[11:13]=="_1"]
    return filenames, seriesnames
    




import pytest
from pandas import Series, DataFrame
from pandas import DataFrame
from acequia import BroGmwXml

@pytest.fixture
def gmw():
    xmlpath = r'.\data\bro\Grondwaterstandonderzoek BRO\GMW000000041126_IMBRO_A.xml'
    return BroGmwXml.from_file(xmlpath)





from collections import OrderedDict
import warnings
import numpy as np
from pandas import DataFrame, Series
import pandas as pd

from .. import gwseries
from .utils import maxfrq, ts1428

def gwtimestats(ts,ref=None,name=None):
    """Return table of groundwater head time series statistics

    Parameters
    ----------
    ts : pd.Series, aq.GwSeries
        Groundwater head time series
    ref : ['datum','surface','mp'], optional, default 'datum'
        Reference level vfor groundwater heads
    name : str, optional
        Groundwater heads series name

    Returns
    -------
    pd.DataFrame
    """

    gwstats = GwTimeStats(ts, ref=ref, name=name)
    return gwstats.stats()


class GwTimeStats:
    """ Return descriptive statistics of groundwater heads time series

    Parameters
    ----------
    ts : pd.Series, aq.GwSeries
        timeseries with groundwater head measurments

    ref : str, ['datum','surface'], optinal
        reference level for measurements

    name : str, optional
        ground water heads series name

    Examples
    --------
    ts = <valid groundwater heads series of GwSeries object>
    tsr = aq.TimeStats(ts)
    tsr.stats()

    Notes
    -----
    Custom function aq.timestats(ts) returns TimeStats.stats() directly.

    """

    def __init__(self, ts, ref=None, name=None):
        """Return TimeStats object"""

        self._ts = ts
        self._name = name


        if isinstance(self._ts,pd.Series):

            self._heads = self._ts
            if ref is not None:
                msg = (f'heads series is type pd.Series, ref={ref} '
                    f'is ignored')
                warnings.warn(msg)
                self._ref = ref

            if self._name is None: 
                self._name = self._ts.name

        if isinstance(self._ts,gwseries.GwSeries):

            self._ref = self._ts._validate_reference(ref)
            self._heads = self._ts.heads(ref=self._ref)

            if self._name is None:
                self._name = self._ts.name()

        if self._name is None:
            self._name = 'series'


    def stats(self):
        """Return time series desciptive statistics

        Returns
        -------
        pd.Dataframe

        """

        heads = self._heads
        stats = Series(name=self._name,dtype='object')

        if not heads.empty:
            stats['firstdate'] = heads.index.min().date()
            stats['lastdate'] = heads.index.max().date()
            stats['minyear'] = heads.index.min().year
            stats['maxyear'] = heads.index.max().year
            stats['yearspan'] = stats['maxyear']-stats['minyear']+1
            stats['nyears'] = len(set(heads.index.year))
            stats['maxfrq'] = maxfrq(heads)
            stats['mean'] = round(heads.mean(),2)
            stats['median'] = round(heads.median(),2)
            q05 = heads.quantile(q=0.05)
            q95 = heads.quantile(q=0.95)
            stats['q05'] = round(q05,2)
            stats['q95'] = round(q95,2)
            stats['dq0595'] = round(q95-q05,2)

        else:

            stats['firstdate'] = np.nan
            stats['lastdate'] = np.nan
            stats['minyear'] = np.nan
            stats['maxyear'] = np.nan
            stats['yearspan'] = np.nan
            stats['nyears'] = np.nan
            stats['maxfrq'] = np.nan
            stats['mean'] = np.nan
            stats['median'] = np.nan
            q05 = np.nan
            q95 = np.nan
            stats['q05'] = np.nan
            stats['q95'] = np.nan
            stats['dq0595'] = np.nan

        self._stats = stats
        return self._stats




""" Funtions for working with files and directories """
import warnings
import os.path
import errno
import os


def listdir(dirpath,filetype=None):
    """Return list of full path of all files in a directory

    Parameters
    ----------
    dirpath : str
        path to directory

    filetype : str, optionally
        list files of type <filetype> only

    Returns
    -------
    list
    """
    if not os.path.isdir(dirpath):
        msg = f'directory {dirpath} not found'
        raise ValueError(msg)

    ##filelist = [f for f in os.listdir(dirpath)
    ##if os.path.isfile(os.path.join(dirpath,f))]
    filelist = [f.path for f in os.scandir(dirpath) if f.is_file()]

    if filetype is not None:
        filelist = [f for f in filelist if f.endswith(filetype)]

    return filelist


def cleardir(dirpath,filetype=None):
    """ delete all existing files in directory 

    Parameters
    ----------
    dirpath : str
        path to directory with files

    filetype : str, optional
        type of files to delete

    """
    for filename in os.listdir(dirpath):

        fpath = os.path.join(dirpath, filename)
        try:
            if os.path.isfile(fpath) or os.path.islink(fpath):
                if filetype is None:
                    os.unlink(fpath)
                elif filetype in os.path.splitext(filename)[1]:
                    os.unlink(fpath)
            #elif os.path.isdir(fpath):
            #    shutil.rmtree(fpath)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))




"""
Module with functions for retrieving data fro BRO REST service.
"""
import datetime as dt
import requests
import lxml.etree as ET
from pandas import Series, DataFrame
import pandas as pd

from .brogldxml import BroGldXml

def _parse_dispatchDocument(tree):
    """"Helper function for parsing BRO XML tree.
    
    Parameters
    ----------
    tree : lxml tree
        XML tree with well properties.

    Returns
    -------
    pd.DataFrame
    """

    nsmap = tree.nsmap
    root = tree.getroottree()

    # dataFrame node properties
    tag = 'dispatchDocument'
    nodes = root.findall(f'.//{tag}',nsmap)
    reclist = []
    for node in nodes:
        eldict = {}
        for el in node.iterdescendants(): # iter all descendants
            tag = ET.QName(el).localname
            eldict[tag]=el.text
        reclist.append(eldict)

    return DataFrame(reclist)



def get_area_wellprops(center=None,radius=None,lowerleft=None,
    upperright=None,startdate='2001-01-01',enddate=None):
    """
    Return properties of wells within an area (rectangle or circle).
    
    Parameters
    ----------
    centre : tuple | list
        Area centre as (latitude,longitude) tuple.
    radius : float
        Area radius in kilometers.
    lowerleft : tuple | list
        Lowerleft corner of rectangle as (latitude,longitude).
    uperright : tuple | list
        Upperright corner of rectangle as (latitude,longitude).
    startdate : str, default '2001-01-01'
        Minimum value of well registration date.
    enddate : str, default today
        Maximum value of well registration date. 

    Returns
    -------
    pd.DataFrame

    Examples
    --------
    > wells = brorest.get_area_wellprops(center=(52.386449,6.919354),radius=0.5)

    """

    # validate startdate
    try:
        date = dt.datetime.strptime(startdate, '%Y-%m-%d')
    except ValueError as e:
        warnings.warn((f'Invalid startdate {startdate} was given. '
            f'Default startdate "2001-01-01" will be used.'))
        date = '2001-01-01'

    # validate enddate
    today = dt.date.today().strftime("%Y-%m-%d")
    if enddate is None:
        enddate = today
    try:
        date = dt.datetime.strptime(enddate, '%Y-%m-%d')
    except ValueError as e:
        warnings.warn((f'Invalid startdate {startdate} was given. '
            f'Wells registered until until today will be selected.'))
        enddate = today

    # circle with radius
    if (center is not None) and (radius is not None):
        lat = center[0]
        lon = center[1]
        json_data = {
            'registrationPeriod': {
                'beginDate': startdate,
                'endDate': enddate,
                },
            'area': {
                'enclosingCircle': {
                    'center': {
                        'lat': lat, #52.349968,
                        'lon': lon, #7.064451,
                        },
                    'radius': radius, #0.5,
                    },
                },
            }

    # rectangle
    if (lowerleft is not None) and (upperright is not None):

        lc_lat = lowerleft[0]
        lc_lon = lowerleft[1]
        uc_lat = upperright[0]
        uc_lon = upperright[1]

        json_data = {
            'registrationPeriod': {
                'beginDate': '2017-01-01',
                'endDate': '2021-01-01',
                },
            'area': {
                'boundingBox': {
                    'lowerCorner': {
                        'lat': lc_lat, #52.340333,
                        'lon': lc_lon, #6.865430,
                        },
                    'upperCorner': {
                        'lat': uc_lat, #52.347915,
                        'lon': uc_lon, #6.888625,
                        },
                    },
                },
            }


    # make request
    headers = {'accept': 'application/xml',}
    params = {'requestReference': 'Punthuizen',}

    response = requests.post('https://publiek.broservices.nl/gm/gmw/v1/characteristics/searches', 
        params=params, headers=headers, json=json_data)
        
    # get xmltree from response
    tree = ET.fromstring(response.content)

    # parse all well nodes
    wellprops = _parse_dispatchDocument(tree)

    return wellprops.drop_duplicates().reset_index(drop=True) # result contains hundreds of duplicates


def get_wellprops(gmwid=None,description=None):
    """Get well properties for one well.
    
    Parameters
    ----------
    gmwid : str
        Valid BRO well id.
    description : str, optional
        User defined description.

    """
    if gmwid is None:
        gmwid = 'GMW000000041033' # for testing
        warnings.warn((f'No BRO well id was given. Values for well '
            f'{gmwid} will be returned.'))

    if description is None:
        description = 'no user description was given'

    headers = {
        'accept': 'application/xml',
        }
    params = {
        'fullHistory': 'ja',
        'requestReference': description,
        }
    response = requests.get(f'https://publiek.broservices.nl/gm/gmw/v1/objects/{gmwid}',
        params=params, headers=headers)

    tree = ET.fromstring(response.content)
    tbl = _parse_dispatchDocument(tree)
    return tbl


def get_welltubes(gmwid):
    """Return well tube number, well tube gldid and instantie for all
    tubes in a well.

    Parameters
    ----------
    gmwid : str
        Valid BRO groundwater monitoring well id.

    Returns
    -------
    pd.DataFrame
        Table with well tube properties.
    """

    # make request
    url = f'https://publiek.broservices.nl/gm/v1/gmw-relations/{gmwid}'
    response = requests.get(url)
    resdict = response.json()

    # iterate over nested json dictionary:
    tubes = []
    for tube in resdict['monitoringTubeReferences']:
        for gld in tube['gldReferences']:
            tubes.append({
                'gmwid':resdict['gmwBroId'],
                'tubenr':tube['tubeNumber'],
                'gldid':gld['broId'],
                'instantie':gld['accountableParty'],
                },)
    return DataFrame(tubes)


def get_putcode(gmwid):
    """Return BRO well user name putcode.

    Parameters
    ----------
    gmwid : str
        Valid BRO groundwater monitoring well id.

    Returns
    -------
    str
    """
    url = ((f'https://publiek.broservices.nl/gm/gmw/v1/well-code/{gmwid}'
        f'?requestReference=myref'))
    response = requests.get(url)
    return response.text


def get_levels(brogld=None,startdate='1900-01-01',enddate=None,reference=None):
    """Return Groundwater Level Data (GLD) for GLD.

    Parameters
    ----------
    brogld : str
        Valid BroGldId.
    startdate : str, default '1900-01-01'
        Start date of groundwater level data.
    enddate : str, default today
        End date of ground water level data.
    reference : str, optional
        Optional user reference for data request.

    Returns
    -------
    BroGldXml
    """
    #brogldid = 'GLD000000009526' #'GLD000000009602'
    if reference is None:
        reference = 'no user reference given' #'Mijn-object-aanvraag-000001'
    if enddate is None:
        enddate = pd.Timestamp.today().strftime('%Y-%m-%d')
    filtered = 'NEE'

    url = ((f'https://publiek.broservices.nl/gm/gld/v1/objects/{brogld}?'
        f'filtered={filtered}&observationPeriodBeginDate={startdate}&'
        f'observationPeriodEndDate={enddate}&requestReference={reference}'))
    response = requests.get(url)

    status_code = response.status_code
    response_string = response.content
    tree = ET.fromstring(response.content)

    # Parse response with BroGldXml
    gld = BroGldXml.from_rest(tree.getroottree())

    return gld


def get_gld_codes(bronhouder):
    """Return list of groundwater level data (GLD) codes for given bronhouder.

    Parameters
    ----------
    bronhouder : str, int
        Valid BRO bronhouder ID number (KvK number).

    Returns
    -------
    list
    """
    bronhouder = str(bronhouder)
    url = f'https://publiek.broservices.nl/gm/gld/v1/bro-ids?bronhouder={bronhouder}'
    response = requests.get(url)
    response_status_code = response.status_code
    return response.json()['broIds']


def get_gmw_codes(bronhouder):
    """Return list of groundwater monitoring well (GMW) codes for given bronhouder.

    Parameters
    ----------
    bronhouder : str, int
        Valid BRO bronhouder ID number (KvK number).

    Returns
    -------
    list
    """
    bronhouder = str(bronhouder)
    url = f'https://publiek.broservices.nl/gm/gmw/v1/bro-ids?bronhouder={bronhouder}'
    response = requests.get(url)
    response_status_code = response.status_code
    return response.json()['broIds']





import pytest
from pandas import DataFrame
from acequia import KnmiDownload
from acequia.read.knmi_download import knmilocations

def test__request_weather():
    """Request weather date from server""" 
    stn = KnmiDownload()
    res = stn._request_weather(par=None)
    assert res.status_code==200

def test__request_precipitation():
    """Request weather date from server""" 
    stn = KnmiDownload()
    res = stn._request_precipitation(par=None)
    assert res.status_code==200

def test_download_with_weather():
    stn = KnmiDownload()
    data = stn.download(kind='weather',result='data')
    assert isinstance(data,DataFrame)
    text = stn.download(kind='weather',result='text')
    assert isinstance(text,str)

def test_download_with_prc():
    stn = KnmiDownload()
    data = stn.download(kind='prc')
    assert isinstance(data,DataFrame)
    text = stn.download(kind='prc',result='text')
    assert isinstance(text,str)

def test_wtr_stns():
    stn = KnmiDownload()
    wtr = stn.wtr_stns
    assert isinstance(wtr,DataFrame)
    assert not wtr.empty

def test_prc_stns():
    stn = KnmiDownload()
    prc = stn.prc_stns
    assert isinstance(prc,DataFrame)
    assert not prc.empty

def test_function_knmi_locations():
    prc = knmilocations(stntype='prc')
    assert isinstance(prc,DataFrame)
    assert not prc.empty

    wtr = knmilocations(stntype='wtr')
    assert isinstance(wtr,DataFrame)
    assert not wtr.empty




import pytest
import numpy as np
import pandas as pd
import acequia as aq

dinodir = '.\\data\\dinogws\\'
jsondir = '.\\output\\json\\'
csvdir = '.\\output\\csv\\'
figdir = '.\\output\\fig\\'
dnpath = f'{dinodir}B29A0850002_1.csv'


@pytest.fixture
def gws():
    return aq.GwSeries.from_dinogws(dnpath)


@pytest.fixture
def name(gws):
    return gws.name()


def test_GwSeries_init():
    gw = aq.GwSeries()
    assert isinstance(gw,aq.gwseries.GwSeries)

def test_len(gws):
    assert len(gws)!=0

def test_repr(gws):
    assert isinstance(repr(gws),str)

def test_GwSeries_from_dinogws(name):
    gwd = aq.GwSeries.from_dinogws(dnpath)
    assert gwd.name()==name


def test_GwSeries_to_json(gws):
    gws.to_json(f'{jsondir}{gws.name()}.json')
    gws.to_json(jsondir)

def test_GwSeries_from_json(gws,name):
    gwj = aq.GwSeries.from_json(f'{jsondir}{gws.name()}.json')
    assert gwj.name()==name


def test_GwSeries_to_csv(gws):
    gws.to_csv(f'{csvdir}{gws.name()}.csv')


def test_GwSeries_name(gws,name):
    assert gws.name()==name


def test_GwSeries_locname(gws):
    assert isinstance(gws.locname(),str)


def test_GwSeries_locprops(gws):
    assert gws.locprops().index[0]==gws.locname()


def test_GwSeries_tubeprops(gws,name):
    assert gws.tubeprops().iloc[0,0]==name
    assert len(gws.tubeprops(last=True))==1


def test_GwSeries_tubeprops_changes(gws):
    assert isinstance(gws.tubeprops_changes(),pd.Series)
    assert not gws.tubeprops_changes().empty


def test_GwSeries_surface(gws):
    assert isinstance(gws.surface(),np.float64)


def test_GwSeries_heads(gws):
    assert isinstance(gws.heads(),pd.Series)
    assert not gws.heads().empty


def test_GwSeries_timestats(gws):
    assert isinstance(gws.timestats(),pd.Series)
    assert not gws.timestats().empty


def test_GwSeries_descibe(gws):
    assert isinstance(gws.describe(),pd.Series)
    assert not gws.describe().empty


def test_GwSeries_plotheads(gws):
    figpath = f'{figdir}plotheads.jpg'
    gws.plotheads(proptype='mplevel',filename=figpath)


def test_GwSeries_gxg(gws):
    gxgs = gws.gxg(ref='surface')
    gxgd = gws.gxg(ref='datum')
    assert isinstance(gxgs,pd.Series)
    assert isinstance(gxgd,pd.Series)
    assert not gxgs.empty
    assert not gxgd.empty
    assert gxgs['gxgref']=='surface'
    assert gxgd['gxgref']=='datum'
    assert gxgs['gt']==gxgd['gt']


def test_GwSeries_xg(gws):
    assert not gws.xg().empty





import pytest
#import numpy as np
from pandas import DataFrame, Series
#import pandas as pd
#import acequia as aq
from acequia import KnmiWeather

fpath = r'.\data\knmi_weather\etmgeg_251.txt' 

@pytest.fixture
def wtr():
    return KnmiWeather(filepath=fpath)

def test_repr(wtr):
    print(wtr)

def test_rawdata(wtr):
    assert isinstance(wtr.rawdata,DataFrame)
    assert not wtr.rawdata.empty

def test_data(wtr):
    assert isinstance(wtr.data,DataFrame)
    assert not wtr.data.empty

def test_meta(wtr):
    assert isinstance(wtr.meta,DataFrame)
    assert not wtr.meta.empty

def test_stn(wtr):
    assert isinstance(wtr.stn,int)

def test_units(wtr):
    assert isinstance(wtr.units,DataFrame)
    assert not wtr.units.empty

def test_timeseries(wtr):
    assert isinstance(wtr.timeseries(),Series)
    assert not wtr.timeseries().empty
    assert isinstance(wtr.timeseries(var='prc'),Series)
    assert not wtr.timeseries(var='prc').empty
    assert isinstance(wtr.timeseries(var='evp'),Series)
    assert not wtr.timeseries(var='evp').empty
    assert isinstance(wtr.timeseries(var='rch'),Series)
    assert not wtr.timeseries(var='rch').empty

def test_prc(wtr):
    assert isinstance(wtr.prc,Series)
    assert not wtr.prc.empty

def test_evp(wtr):
    assert isinstance(wtr.evp,Series)
    assert not wtr.evp.empty

def test_recharge(wtr):
    assert isinstance(wtr.recharge,Series)
    assert not wtr.recharge.empty



""" Conversion of geographic coordinates


"""
import numpy as np

class CrdCon:
    """Convert coordinates from Dutch RD Grid to WGS84 and back

    Methods
    -------
        RDtoWGS84(X,Y,Zone=False)
            convert x,y in RD to Lon,Lat in WGS84
        WGS84toRD(Lon,Lat)
            convert Lon,Lat in WGS84 to x,y in RD

    Examples
    --------
    >> crc = acequia.CrdCon()
    >> crc.RDtoWGS84(233883.131, 82065.167)
    >> {'Lon': 6.458551281201178,
        'Lat': 48.726255249201174,
        'East': 313102.7205129704,
        'North': 5400142.229150799,
        'xRD': 233883.131,
        'yRD': 82065.167,
        'UMTZONE': 'UMT32'}
    >> crc.WGS84toRD(4.88352,52.3745)
    >> {'xRD': 233883.06517103617,
        'yRD': 582067.0401935352,
        'Lon': 6.5682,
        'Lat': 53.2194}

    Note
    ----
    Coordinate onversions are valid only for the Netherlands!
    
    Formulas from article by Schreutelkamp en Strang Van Hees:
    "Benaderingsformules voor de transformatie tussen RD-en 
     WGS84-kaartcoordinaten"

    link: http://www.dekoepel.nl/pdf/Transformatieformules.pdf

    """

    Towers = {
             "Westertoren": {
                "town": "Amsterdam",
                "xRD": 120700.723,
                "yRD": 487525.501,
                "Lat_WGS84": 52.3745311,
                "Lon_WGS84": 4.8835223,
                "Lat_DDMMmm" : ['52','22.472'],
                "Lon_DDMMmm" : ['04','53.011'],
                "Lat_DDMMSS" : ['52','22','28.3'],
                "Lon_DDMMSS" : ['04','53','0.7'],
                "East_UMT31": 628217.312,
                "North_UMT31": 5804365.552,
                "East_UMT32": 219827.625, #calculated from RD
                "North_UMT32": 5810673.509, #calculated from RD
                },
            "Martinitoren": {
                "town": "Groningen",
                "xRD": 233883.131,
                "yRD": 582065.167,
                "Lat_WGS84": 53.2193814,
                "Lon_WGS84": 6.5682021,
                "Lat_DDMMmm" : ['53','13.163'],
                "Lon_DDMMmm" : ['06','34.092'],
                "Lat_DDMMSS" : ['53','13','9.8'],
                "Lon_DDMMSS" : ['06','34','5.5'],
                "East_UMT31": 738204.153, #calculated from RD
                "North_UMT31": 5902619.635, #calculated from RD
                "East_UMT32": 337643.235,
                "North_UMT32": 5899435.841,}
            }


    X0 = 155000.00
    Y0 = 463000.00
    Phi0 = 52.15517440
    Lambda0 = 5.38720621
    E0zone31 = 663304.11
    N0zone31 = 5780984.54
    E0zone32 = 252878.65
    N0zone32 = 5784453.44

    def RDtoWGS84(self,X,Y,Zone=False):
        """ Convert RD Xcoor,Ycoor to WGS84 latitude,longitude
        
        Parameters
        ----------
        X : float
            x-coordinate in RD-system
        Y : float
            y-coordinate in RD-system
        Zone : bool, optinal
            Apply different formulas for easting below 6 (UTM31) and
            eastignabove 6 (UTM32). Default is True.

        Returns
        -------
        dict

        Example
        -------
        >> crc = acequia.CrdCon()
        >> crc.RDtoWGS84(233883.131, 82065.167)
        >> {'Lon': 6.458551281201178,
            'Lat': 48.726255249201174,
            'East': 313102.7205129704,
            'North': 5400142.229150799,
            'xRD': 233883.131,
            'yRD': 82065.167,
            'UMTZONE': 'UMT32'}

        Note
        ----
        Within the WGS84 system, the 6th meridian divides the Netherslands 
        into two a western and an eastern zone. The aerea west of the 6th 
        meridian lies UTM zone 31, the eastern part belong to UTM zone 32. 
        To avoid jumps in transformed WGS84 coordinates, the default 
        'Zone=False' unsures that all transformations are calculated with 
        the paramaters for UTMzone 31.
        This follows the suggestion in the orginale article by Schreutelkamp 
        and Strang van Hees, from which all formules wer copied. 


        """
        Lon = self._RDtoWGS84Lon(X,Y)
        Lat = self._RDtoWGS84Lat(X,Y)
        if Lon>6 and Zone:
            [E,N]=self._RDtoWGS84forUMT32(X,Y)
            UMTZONE = "UMT32"
        else:
            [E,N]=self._RDtoWGS84forUMT31(X,Y)
            UMTZONE = "UMT31"
        return {"Lon" : Lon, "Lat" : Lat, "East" : E, "North" : N, 
                "xRD" : X, "yRD" : Y, "UMTZONE" : UMTZONE}


    def WGS84toRD(self,Lon,Lat):
        """ Convert WGS84 latitude,longitude to RD Xcoor,Ycoor
        
        Parameters
        ----------
        Lon : float
            Longitude in WGS84
        Lat : float
            Latitude in WGS84
        Zone : bool, optinal
            Apply different formulas for easting below 6 (UTM31) and
            easting above 6 (UTM32). Default is True.

        Returns
        -------
        dict

        Example
        -------
        >> crc = acequia.CrdCon()
        >> crc.WGS84toRD(4.88352,52.3745)
        >> {'xRD': 233883.06517103617,
            'yRD': 582067.0401935352,
            'Lon': 6.5682,
            'Lat': 53.2194}

        Note
        ----
        Allows for flexible input format of Lat,Lon:
            DD.ddd       : float or str with decimal degrees (default)
            [DD,MM,SS.S] : list (n=3) of float or str giving
                           degrees, minutes and seconds
            [DD,MM.mmmm] : list (n=2) of float or str giving
                           degrees and decimal minutes


        """

        try:

            if (type(Lon)==str) & (type(Lat)==str):

                Lon = float(Lon)
                Lat = float(Lat)


            elif (type(Lon)==list) & (type(Lat)==list):

                def tofloat(mystring):
                    """
                    if (len(mystring)!=0) & mystring[0]=='0':
                        result = float(mystring[1])
                    else: 
                        result = float(mystring)
                    """
                    try:
                        result = float(mystring)
                    except:
                        print('Could not convert to float: {mystring}')
                        raise
                    return result

                if (len(Lon)==3) & (len(Lat)==3):

                    def DDMMSSs(mylist):
                        """ conversie graden minuten seconden (notatie [DD,MM,SS.S])
                        naar decimale graden (notatie DD.ddd) 
                        Martinitoren: N [52,22,28.3], E [4,53,0.7] wordt N 52.37453, E 4.88352 """
                        dd = tofloat(mylist[0])
                        mm = tofloat(mylist[1])
                        ss = tofloat(mylist[2])
                        return dd+(mm*60+ss)/3600

                    Lon = DDMMSSs(Lon)
                    Lat = DDMMSSs(Lat)

                elif (len(Lon)==2) & (len(Lat)==2):

                    def DDMMMmmm(mylist):
                        """ conversie graden minuten (notatie [DD,MM.mmm])
                        naar decimale graden (notatie DD.ddd) 
                        Martinitoren: N [52,22.472], E [4,53.011] wordt N 52.37453, E 4.88352 """
                        dd = tofloat(mylist[0])
                        mm = tofloat(mylist[1])
                        return dd+(mm*60)/3600

                    Lon = DDMMMmmm(Lon)
                    Lat = DDMMMmmm(Lat)

            xcoor = self._WGS84toRDx(Lon,Lat)
            ycoor = self._WGS84toRDy(Lon,Lat)

        except Exception as err:
            xcoor = None
            ycoor = None
            print("Fout opgetreden: ",err.args)
            raise Warning("Error in WGS842RD")

        finally:
            result = {
                "xRD" : xcoor,
                "yRD" : ycoor,
                "Lon"  : Lon,
                "Lat"  : Lat}
        return result


    def _RDtoWGS84Lon(self,X,Y):
        """Calculate WGS84 Longitude from RD X,Y """

        coef = [
               [1,0,5260.52916],
               [1,1,105.94684],
               [1,2,2.45656],
               [3,0,-0.81885],
               [1,3,0.05594],
               [3,1,-0.05607],
               [0,1,0.01199],
               [3,2,-0.00256],
               [1,4,0.00128],
               [0,2,0.00022],
               [2,0,-0.00022],
               [5,0,0.00026]
               ]
        dX=(X-self.X0)*pow(10,-5)
        dY=(Y-self.Y0)*pow(10,-5)
        Lambda = 0
        for i in range(len(coef)):
            p = coef[i][0]
            q = coef[i][1]
            L = coef[i][2]
            Lambda+=L*pow(dX,p)*pow(dY,q)
        return self.Lambda0+Lambda/3600.0


    def _RDtoWGS84Lat(self,X,Y):
        """Calculate WGS84 Latitude from RD X,Y """

        coef = [
               [0,1,3235.65389],
               [2,0,-32.58297],
               [0,2,-0.24750],
               [2,1,-0.84978],
               [0,3,-0.06550],
               [2,2,-0.01709],
               [1,0,-0.00738],
               [4,0,0.00530],
               [2,3,-0.00039],
               [4,1,0.00033],
               [1,1,-0.00012]
               ]
        dX=(X-self.X0)*pow(10,-5)
        dY=(Y-self.Y0)*pow(10,-5)
        phi = 0
        for i in range(len(coef)):
            p = coef[i][0]
            q = coef[i][1]
            K = coef[i][2]
            phi+=K*pow(dX,p)*pow(dY,q)
        return self.Phi0+phi/3600.0

    def _RDtoWGS84forUMT31(self,X,Y):
        """Convert RD-system X,Y to WGS84 Easting,Northing 
        using paramters for Zone UMT31 """

        A0 = self.E0zone31
        B0 = self.N0zone31
        A1 = 99947.539
        B1 = 3290.106
        A2 = 20.008
        B2 = 1.310
        A3 = 2.041
        B3 = 0.203
        A4 = 0.001
        B4 = 0.000

        dX = (X-self.X0)*pow(10,-5)
        dY = (Y-self.Y0)*pow(10,-5)

        E = A0 + A1*dX - B1*dY + A2*(pow(dX,2)-pow(dY,2)) - B2*(2*dX*dY)
        E+= A3*(pow(dX,3)-3*dX*pow(dY,2))-B3*(3*pow(dX,2)*dY-pow(dY,3))
        E+= A4*(pow(dX,4)-6*pow(dX,2)*pow(dY,2)+pow(dY,4))-B4*(4*pow(dX,3)*dY-4*pow(dY,3)*dX)

        N = B0 + B1*dX + A1*dY + B2*(pow(dX,2)-pow(dY,2)) + A2*(2*dX*dY)
        N+= B3*(pow(dX,3)-3*dX*pow(dY,2))+A3*(3*pow(dX,2)*dY-pow(dY,3))
        E+= B4*(pow(dX,4)-6*pow(dX,2)*pow(dY,2)+pow(dY,4))+A4*(4*pow(dX,3)*dY-4*pow(dY,3)*dX)
        return [E,N]

    def _RDtoWGS84forUMT32(self,X,Y):
        """Convert RD-system X,Y to WGS84 Easting,Northing 
        using paramters for Zone UMT32 """

        A0 = self.E0zone32
        B0 = self.N0zone32
        A1 = 99919.783
        B1 = -4982.166
        A2 = -30.208
        B2 = 3.016
        A3 = 2.035
        B3 = -0.309
        A4 = -0.002
        B4 = 0.001

        dX = (X-self.X0)*pow(10,-5)
        dY = (Y-self.Y0)*pow(10,-5)

        E = A0 + A1*dX - B1*dY + A2*(pow(dX,2)-pow(dY,2)) - B2*(2*dX*dY)
        E+= A3*(pow(dX,3)-3*dX*pow(dY,2))-B3*(3*pow(dX,2)*dY-pow(dY,3))
        E+= A4*(pow(dX,4)-6*pow(dX,2)*pow(dY,2)+pow(dY,4))-B4*(4*pow(dX,3)*dY-4*pow(dY,3)*dX)

        N = B0 + B1*dX + A1*dY + B2*(pow(dX,2)-pow(dY,2)) + A2*(2*dX*dY)
        N+= B3*(pow(dX,3)-3*dX*pow(dY,2))+A3*(3*pow(dX,2)*dY-pow(dY,3))
        E+= B4*(pow(dX,4)-6*pow(dX,2)*pow(dY,2)+pow(dY,4))+A4*(4*pow(dX,3)*dY-4*pow(dY,3)*dX)
        return [E,N]


    def _WGS84toRDx(self,Lon,Lat):
        """Convert WGS longitude,latitude to RD-system Xcrd """

        Lambda = Lon
        Phi = Lat

        coef = [#p,q,Rpg
               [0,1,190094.945],
               [1,1,-11832.228],
               [2,1,-114.221],
               [0,3,-32.391],
               [1,0,-0.705],
               [3,1,-2.340],
               [1,3,-0.608],
               [0,2,-0.008],
               [2,3,0.148]
               ]
        dPhi=0.36*(Phi-self.Phi0)
        dLambda=0.36*(Lambda-self.Lambda0)
        X = 0
        for i in range(len(coef)):
           p = coef[i][0]
           q = coef[i][1]
           R = coef[i][2]
           X+=R*pow(dPhi,p)*pow(dLambda,q)
        return self.X0+X


    def _WGS84toRDy(self,Lon,Lat):
        """Convert WGS longitude,latitude to RD-system Ycrd """

        Lambda = Lon
        Phi = Lat
        coef = [
               [1,0,309056.544],
               [0,2,3638.893],
               [2,0,73.077],
               [1,2,-157.984],
               [3,0,59.788],
               [0,1,0.433],
               [2,2,-6.439],
               [1,1,-0.032],
               [0,4,0.092],
               [1,4,-0.054]
               ]
        dPhi=0.36*(Phi-self.Phi0)
        dLambda=0.36*(Lambda-self.Lambda0)
        Y = 0
        for i in range(len(coef)):
            p = coef[i][0]
            q = coef[i][1]
            S = coef[i][2]
            Y+=S*pow(dPhi,p)*pow(dLambda,q)
        return self.Y0+Y


    def _WGS84toRDforUMT31(self,E,N):
        """ Convert WGS84 Easting, Northing tot RD-system X,Y-self
        using paramters for Zone UMT31 """

        C0 = self.X0
        D0 = self.Y0
        C1 = 99944.187
        D1 = -3289.996
        C2 = -20.039
        D2 = 0.668
        C3 = -2.042
        D3 = 0.066
        C4 = 0.001
        D4 = 0.000

        dE = (E-self.E0zone31)*pow(10,-5)
        dN = (N-self.N0zone31)*pow(10,-5)

        X = C0 + C1*dE - D1*dN + C2*(pow(dE,2)-pow(dN,2))-D2*(2*dE*dN)
        X+= C3*(pow(dE,3)-3*dE*pow(dN,2))-D3*(3*pow(dE,2)*dN-pow(dN,3))
        X+= C4*(pow(dE,4)-6*pow(dE,2)*pow(dN,2)+pow(dN,4))-D4*(4*pow(dE,3)*dN-4*pow(dN,3)*dE)

        Y = D0 + D1*dE + C1*dN + D2*(pow(dE,2)-pow(dN,2))+C2*(2*dE*dN)
        Y+= D3*(pow(dE,3)-3*dE*pow(dN,2))+C3*(3*pow(dE,2)*dN-pow(dN,3))
        Y+= D4*(pow(dE,4)-6*pow(dE,2)*pow(dN,2)+pow(dN,4))+C4*(4*pow(dE,3)*dN-4*pow(dN,3)*dE)

        return [X,Y]


    def _WGS84toRDforUMT32(self,E,N):
        """ Convert WGS84 Easting, Northing tot RD-system X,Y-self
        using paramters for Zone UMT32 """

        C0 = 155000.00
        D0 = 463000.00
        C1 = 99832.079
        D1 = 4977.793
        C2 = 30.280
        D2 = 1.514
        C3 = -2.034
        D3 = -0.099
        C4 = -0.001
        D4 = 0.000

        dE = (E-self.E0zone32)*pow(10,-5)
        dN = (N-self.N0zone32)*pow(10,-5)

        X = C0 + C1*dE - D1*dN + C2*(pow(dE,2)-pow(dN,2))-D2*(2*dE*dN)
        X+= C3*(pow(dE,3)-3*dE*pow(dN,2))-D3*(3*pow(dE,2)*dN-pow(dN,3))
        X+= C4*(pow(dE,4)-6*pow(dE,2)*pow(dN,2)+pow(dN,4))-D4*(4*pow(dE,3)*dN-4*pow(dN,3)*dE)

        Y = D0 + D1*dE + C1*dN + D2*(pow(dE,2)-pow(dN,2))+C2*(2*dE*dN)
        Y+= D3*(pow(dE,3)-3*dE*pow(dN,2))+C3*(3*pow(dE,2)*dN-pow(dN,3))
        Y+= D4*(pow(dE,4)-6*pow(dE,2)*pow(dN,2)+pow(dN,4))+C4*(4*pow(dE,3)*dN-4*pow(dN,3)*dE)

        return [X,Y]





import pytest
import acequia as aq

dinodir = '.\\data\\dinogws\\'
hymonpath = '.\\data\\hymon\hydromonitor_testdata.csv'

def test_GwList_init():

    with pytest.raises(ValueError):
        # arguments srcdir and srcfile both missing
        gwl = aq.GwList()

    with pytest.raises(ValueError):
        # invalid srctype
        gwl = aq.GwList(srcdir=dinodir,srctype='dummy')
    
    with pytest.raises(ValueError):
        # srcdir does not exist
        gwl = aq.GwList(srcdir='dummy')





import pytest
import lxml.etree as ET
from pandas import Series, DataFrame
from pandas import DataFrame
from geopandas import GeoDataFrame
from acequia import BroGldXml


@pytest.fixture
def filegld():
    xmlpath = r'.\data\bro\Grondwaterstandonderzoek BRO\GLD000000009526_IMBRO_A.xml'
    return BroGldXml.from_file(xmlpath)

@pytest.fixture
def restgld():
    # parse xml file that contains an xml from the BRO REST service
    xmlpath = r'.\data\bro\RESTXML\GLD000000009526.xml'  #GLD000000010071.xml'
    tree = ET.parse(xmlpath)
    return BroGldXml.from_rest(tree)

@pytest.mark.parametrize('gld', [filegld, restgld])
def test_property_gldprops(gld,request):
    gld = request.getfixturevalue(gld.__name__)
    assert isinstance(gld.gldprops,Series)
    assert not gld.gldprops.empty

@pytest.mark.parametrize('gld', [filegld, restgld])
def test_property_obs(gld,request):
    gld = request.getfixturevalue(gld.__name__)
    assert isinstance(gld.obs,DataFrame)
    assert not gld.obs.empty

@pytest.mark.parametrize('gld', [filegld, restgld])
def test_property_obsprops(gld,request):
    gld = request.getfixturevalue(gld.__name__)
    assert isinstance(gld.obsprops,DataFrame)
    assert not gld.obsprops.empty

@pytest.mark.parametrize('gld', [filegld, restgld])
def test_property_procesprops(gld,request):
    gld = request.getfixturevalue(gld.__name__)
    assert isinstance(gld.procesprops,DataFrame)
    assert not gld.procesprops.empty

@pytest.mark.parametrize('gld', [filegld, restgld])
def test_property_heads(gld,request):
    gld = request.getfixturevalue(gld.__name__)
    assert isinstance(gld.heads,Series)
    assert not gld.heads.empty




import pytest
from datetime import datetime
from pandas import DataFrame, Series
from acequia import DinoGws, GwSeries
from acequia.read.dinogws import filesfromdir

fpath = r'.\data\dinogws\B28A0475002_1.csv'
fdir = r'.\data\dinogws\\'

@pytest.fixture
def dn():
    dn = DinoGws(filepath=fpath,readall=True)
    assert isinstance(dn,DinoGws)
    return dn

def test_repr(dn):
    assert isinstance(str(dn),str)

def test_parse_dino_date(dn):
    assert isinstance(dn.parse_dino_date('17-08-2009'),datetime)

def test_get_heads(dn):
    assert isinstance(dn.get_heads(),Series)
    assert not dn.get_heads().empty
    assert isinstance(dn.get_heads(units='cmmv'),Series)
    assert not dn.get_heads(units='cmmv').empty
    assert isinstance(dn.get_heads(units='cmmp'),Series)
    assert not dn.get_heads(units='cmmp').empty
    assert isinstance(dn.get_heads(units='cmnap'),Series)
    assert not dn.get_heads(units='cmnap').empty
    
def test_headdata(dn):
    assert isinstance(dn.headdata,DataFrame)
    assert not dn.headdata.empty

def test_header(dn):
    assert isinstance(dn.header,DataFrame)
    assert not dn.header.empty

def test_data(dn):
    assert isinstance(dn.data,DataFrame)
    assert not dn.data.empty

def test_gwseries(dn):
    gw = dn.gwseries
    assert isinstance(gw, GwSeries)
    assert not gw.heads().empty
    assert not gw.tubeprops().empty
    assert not gw.locprops().empty

def test_locname(dn):
    assert isinstance(dn.locname,str)
    assert len(dn.locname)!=0

def test_filname(dn):
    assert isinstance(dn.filname,str)
    assert len(dn.filname)!=0

def test_srname(dn):
    assert isinstance(dn.srname,str)
    assert len(dn.srname)!=0

def test_describe(dn):
    assert isinstance(dn.describe,DataFrame)
    assert not dn.describe.empty

def test_locations(dn):
    assert dn.get_locations().empty

def test_merge(dn):
    assert isinstance(dn,DinoGws)

def test_filesfromdir(fdir=fdir):
    res = filesfromdir(r'.\data\dinogws\\')
    assert isinstance(res,tuple)
    assert len(res)==2


# test private methods

def test__readfile(dn):
    res = dn._readfile(filepath=dn.filepath)
    assert isinstance(res,list)
    assert len(res)!=0

def test__readlines(dn):
    res = dn._readlines()
    assert isinstance(res,tuple)
    assert len(res)==2

def test__findlines(dn):
    res = dn._findlines()
    assert isinstance(res,tuple)
    assert len(res)==3

def test__readheader(dn):
    assert isinstance(dn._readheader(),DataFrame)
    assert not dn._readheader().empty

def test__readgws(dn):
    assert isinstance(dn._readgws(),DataFrame)
    assert not dn._readgws().empty




"""This module contains tools for working with the WaterWeb class"""

from pathlib import Path
import warnings
import pandas as pd
from .waterweb import WaterWeb

def measurement_types(fdir,zeros=True,rowsum=True,colsum=True):
    """
    Return table of measurement types for mutiple networks

    Parameters
    ----------
    fdir : str
        valid directory path with WaterWeb csv sourcefiles
    zeros : bool, default True
        show zeros in table
    rowsum : bool, default True
        add column with row totals
    colsum : bool, default True
        add row with column totals

    Returns
    -------
    pd.DataFrame
        table of measurement type counts for all networks
    """

    if not Path(fdir).exists:
        raise ValueError(f'Directory {fdir} does not exist.')

    # list of series with counts for each network file
    pathlist = Path(fdir).glob('**/*')
    filelist = [x for x in pathlist if x.is_file()]

    counts_list = []
    for fname in filelist:
        wwn = WaterWeb.from_csv(fname)
        counts_list.append(wwn.measurement_types)

    # table of measurment type by network names
    tbl_list = [pd.DataFrame(sr).T for sr in counts_list]
    tbl = pd.concat(tbl_list).fillna(0)
    for col in tbl.columns:
        tbl[col] = tbl[col].astype(int)

    # sort column names
    if not set(tbl.columns) - set(wwn.MEASUREMENT_TYPES):
        tbl = tbl.reindex(wwn.MEASUREMENT_TYPES, axis=1)
    else: #tbl2.columns contains names not in _measurement_types
        warnings.warn('Non-standard measurement types found.')
        tbl = tbl.reindex(sorted(tbl.columns), axis=1)

    if rowsum==True:
        tbl['total'] = tbl.sum(axis=1)

    if colsum==True:
        tbl.loc['total',:] = tbl.sum()

    if not zeros: # show no zeros but empty string
        tbl = tbl.copy()
        for col in tbl.columns:
            tbl[col] = tbl[col].astype(str)
        tbl = tbl.replace('0','')

    return tbl
